/* 
   pair forces by the direct site-site method, optional cache optimization
   #included by cook/forces.c #ifndef LINKCELL 
*/

static vector *Arp,*Brp;

static void onepair(int n,int m)
{
  molecule_t *mn=cfg+n,*mm=cfg+m;

  double Emn = (*pot[mn->sp][mm->sp]) (rof(mn,Brp),rof(mm,Brp),mn,mm,Arp);

  if (m==n) En.intra += Emn;
  else En.pot += Emn;
  En.LJ += En.LJmn;
}

/* recursive */
static void subDivide(int x,int y, int l)
/*    #
    o o
  o o o
* o o % *=(x,y), %=(x+l,y) #=(x+l,y+l)
*/

{
  if (l==0)
    onepair(x,y);
  else if (abs(l)==1) {
    onepair(x,y);
    onepair(x+l,y);
    onepair(x+l,y+l); }
  else if (abs(l)==2) {
    /* unnecessary optimization */
    int lh=l/2;

    onepair(x,y);
    onepair(x+lh,y);
    onepair(x+lh,y+lh);
    onepair(x+l,y);
    onepair(x+l,y+lh);
    onepair(x+l,y+l); }
  else {
    int lh=l/2,sg=l>0?1:-1,lh1=lh+sg,lh2=l-lh1;

    subDivide(x,y,lh);
    subDivide(x+lh1,y,lh2);
    subDivide(x+lh1,y+lh1,lh2);
    if (l&1) subDivide(x+l-sg,y+lh,sg-lh2);
    else     subDivide(x+l,y+lh,-lh2); }
}

static void pairforces(ToIntPtr B, ToIntPtr A) /***************** pairforces */
{
  int n,m,sp;
  vector *f,*rp=A->rp;
#ifdef Hamilton
  vector *p,*dr,*dp,
#endif /*# Hamilton */
  molecule_t *mn,*mm;
  double Emn;
  double Enel=En.el;

  if (cache<=0 && (No.first || option('j')>0)) {
    WARNING(("cache=0 (recursive triangulation) does not support:\n\
*** monitoring energy of the 1st molecule (No.first)\n\
*** fixed positions of several first molecules (option -j)\n\
*** => cache=64 set instead"))
      cache=64; }

  if (cache<=0) {
    Arp=A->rp;
    Brp=B->rp;
    /* recursive triangulation, not good for small molecules */
    subDivide(0,0,No.N-1); }
  else if (cache==1) loop (n,FROM,No.N) {
    /* simplest code */
    /*** pair forces and real part of Ewald sums ***/
    /*** in the Hamilton formalism this loop may be merged to the next loop ***/
    mn=cfg+n;
    sp=mn->sp;
#ifdef Lagrange
    f=rof(mn,B->rp);
    loopto (m,0,n) {
      mm=cfg+m;
      Emn = (*pot[sp][mm->sp]) (f,rof(mm,B->rp),mn,mm,rp);
      if (m==n) En.intra += Emn;
      else En.pot += Emn;
      En.LJ += En.LJmn;
      if (No.first) {
        if (n<No.first) {
          En.LJ0+=En.LJmn; En.el0+=(En.el-Enel); En.pot0+=Emn; }
        else if (m<No.first) {
          En.LJX+=En.LJmn; En.elX+=(En.el-Enel); En.potX+=Emn; }
        Enel=En.el; } }
#else /*# Lagrange */
    dp=pof(mn,B->rp);
    loopto (m,0,n) {
      mm=cfg+m;
      Emn = (*pot[sp][mm->sp]) (dp,pof(mm,B->rp),mn,mm,rp);
      if (m==n) En.intra += Emn; else En.pot += Emn; }
#endif /*# Lagrange */
    } /* n */
  else {
    /* cache > 1: systematic blocking */
    int n0,nf,nt,m0,mt;
#ifdef Hamilton
    ERROR(("cache + Hamilton"))
#endif
    /*** pair forces and real part of Ewald sums ***/
    /*** in the Hamilton formalism this loop may be merged to the next loop ***/
    for (n0=0; n0<No.N; n0+=cache) {
      nf=max(FROM,n0);
      nt=min(No.N,n0+cache); /* not incl. */
      for (m0=0; m0<=n0; m0+=cache) {
        loop (n,nf,nt) {
          mt=min(m0+cache-1,n);
          mn=cfg+n;
          sp=mn->sp;
          f=rof(mn,B->rp);
          loopto (m,m0,mt) {
            mm=cfg+m;
            Emn = (*pot[sp][mm->sp]) (f,rof(mm,B->rp),mn,mm,rp);
            if (m==n) En.intra += Emn;
            else En.pot += Emn;
            En.LJ += En.LJmn;
            if (No.first) {
              if (n<No.first) {
                En.LJ0+=En.LJmn; En.el0+=(En.el-Enel); En.pot0+=Emn; }
              else if (m<No.first) {
                En.LJX+=En.LJmn; En.elX+=(En.el-Enel); En.potX+=Emn; }
              Enel=En.el; } } } /* m,n */
      } } /* m0,n0 */
    } /* cache */

  if (option('v')&4) if (measure)
    prt("after mol-mol: intramol=%g intermol=%g En.el(sum)=%g",
        En.intra,En.pot,En.el);

  //  fprintf(stderr,"PAR %g %08p: %g %g %g\n",t,A,A->rp[900][0],A->rp[900][1],A->rp[900][2]);

  En.pot += En.intra;
}
