/*
module - header file = "rndgen.h"

random number generator - simple and portable generator
^^^^^^^^^^^^^^^^^^^^^^^   good enough for initializing MD or similar
                          rnd() in (0,1), grid ~2^-31
                          Don't use for MC integration and simulations
                          (serious correlations of rare events)!

initialization: rndinit(tablen,seed)
where tablen = irrelevant
      seed = any integer (0 uses seed from time)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h> /* fmod() */

void extError(const char *msg, const char *f, int l);
#define Error(X) extError(X,__FILE__,__LINE__)

#include "rndgen.h"
#include "rndseed.c"

struct { double seed; } rndgen;

unsigned4 Irnd(void)
/* random number in [0,2147483647): inefficiently from double */
{
  rndgen.seed=fmod(rndgen.seed*16807e0,2147483647e0);
  return (unsigned4)rndgen.seed;
}

double rnd(void)
/* random number in [0,1) */
{
  rndgen.seed=fmod(rndgen.seed*16807e0,2147483647e0);
  return rndgen.seed/2147483647e0;
}

double rndcos(void)
/* random number in [-1,1) */
{
  rndgen.seed=fmod(rndgen.seed*16807e0,2147483647e0);
  return rndgen.seed/1073741823.5-1;
}

unsigned4 irnd(unsigned N)
{
  rndgen.seed=fmod(rndgen.seed*16807e0,2147483647e0);
  return (int)(rndgen.seed/2147483647e0*N);
}

#include "rndetc.c"

char rndinfo[80]="rndgen0: *7^5 mod 2^31, grid=4.6566e-10";

unsigned4 rndinit(int4 tablen,int4 seed)
/* returns max possible Irnd() */
{
 if (seed==0) rndgen.seed=seed_from_time();
 else rndgen.seed=abs(seed);
 if (rndgen.seed==0) rndgen.seed=1;

 sprintf(rndinfo+39," seed=%.0f",rndgen.seed);
 rnd();

 return (unsigned4)2147483646e0;
}
