/*
  constraint dynamics (Lagrange, Hamilton, Shake)
  global #defines:
    Hamilton (Hamiltonian version of constraint dynamics, obsolete)
    SHAKE: 4 versions or off
    POLAR: selffield calculations
    SHEAR: shear viscosity support
    VERLET: with SHAKE: kin.energy algorithm
  local #defines:
    PRTKIN LAMBDA PRECOND DEBUG
*/

/*
#define DEBUG
  debug energy terms, pressure, scaling */

#include "ground.h"
#define SDS
#include "alloc.h"
#include "statics.h"
#include "simglob.h"
#include "simcg.h"
#include "norm.h"
#include "constrd.h"
#include "forces.h" /* the same for LINKCELL */
#include "cputime.h"
#include "rndgen.h"
#include "units.h"
#include "simils.h" /* Fn() */
#include "asksig.h"
#ifdef LINKCELL
#  include "linklist.h"
#endif /*# LINKCELL */

#ifdef POLAR
#  include "intermac.h"
#  if POLAR&2
#    include "log1.c"
#  endif /*# POLAR&2 */
#  ifndef FREEBC
#    include "elst.h"
#  endif /*# FREEBC */
#endif /*# POLAR */

#ifdef SHEAR
#  include "units.h"
#endif /*# SHEAR */
#ifdef Hamilton
double gsq;       /* sum sqr of Lagr Mult (Hamilton only) */
#endif /*# Hamilton */

#ifdef ANCHOR
void normalizer(vector cm)
/* local version of normalize for 1 vector, see also normalize in norm.c */
{
#  ifndef FREEBC
  if (cm[0]<-box.Lh[0]) cm[0]+=box.L[0]; else if (cm[0]>box.Lh[0]) cm[0]-=box.L[0];
  if (cm[1]<-box.Lh[1]) cm[1]+=box.L[1]; else if (cm[1]>box.Lh[1]) cm[1]-=box.L[1];
#    ifndef WALL
  if (cm[2]<-box.Lh[2]) cm[2]+=box.L[2]; else if (cm[2]>box.Lh[2]) cm[2]-=box.L[2];
#    endif /*# WALL */
#  endif /*# FREEBC */
}
#endif /*# ANCHOR */

#if PRESSURETENSOR&PT_VIR
static double PTvirc[PT_DIM];  /* of constraint forces */
#endif /*# PRESSURETENSOR&PT_VIR */

static double blogT(double E,double T) /****************************** blogT */
/* 1/2*ln(E/kT), but always kept in [-1,1] */
{
  const double exp2=7.38905609893065022723;

  if (E>T*exp2) return 1;
  if (E*exp2<T) return -1;

  return log(E/T)/2;
}

static void measureconstraints(void) /******************* measureconstraints */
{
  int i;
  double V=box.L[0]*box.L[1]*box.L[2];

  StaSet(0,lag.err,2,lag.n);
  StaAdd("Pvirc [Pa]",En.virc*(Punit/DIM)/V);
  StaAdd("Eel [K]",En.el);
  StaAdd("Pvir pair [Pa]",En.vir*(Punit/DIM)/V);
  En.vir+=En.virc+En.el;

#ifdef POLAR
  StaAdd("Eself [K]",En.self);
  /* 9/2001: -En.self*2 added instead of the virial of forces on aux
     charges that incorrectly did not include Ewald corrections.
     WARNING: not tested for axial and saturated polarizability !!! */
  En.vir-=En.self*2; /* see also VVV(En.Pvir,-=rpol[i],*fpol[i]) below */
#endif /*# POLAR */

#if PRESSURETENSOR&PT_KIN
  {
    static int pass=0;
    double x=SUM(En.Pkin)/2-En.kin;

    if (!pass && fabs(x)>1000) {
      prt("WARNING: Ekin=%g K  tr(Pkin)/2=%g K  dif=%g K\n\
          (more warnings suppressed)",
          En.kin,SUM(En.Pkin)/2,x);
      pass=1; }
  }
  loop (i,0,PT_DIM) En.Pkin[i]*=Punit/V; // /DIM removed (BUG in 2.4m)
#  if PRESSURETENSOR&PT_MOL
  {
    static int pass=0;
    double x=SUM(En.PKin)/2-En.kin_tr/2; /* still doubled */

    if (!pass && fabs(x)>1000) {
      prt("WARNING: Ekininter=%g K  tr(PKin)=%g K  dif=%g K\n\
          (more warnings suppressed)",
          En.kin_tr/2,SUM(En.PKin)/2,x);
      pass=1; }
  }
  loop (i,0,PT_DIM) En.PKin[i]*=Punit/V; // /DIM removed
#  endif /*# PRESSURETENSOR&PT_MOL */
#endif /*# PRESSURETENSOR&PT_KIN */

#if PRESSURETENSOR&PT_VIR
  loop (i,0,PT_DIM) {
    En.Pvir[i]+=PTvirc[i];
    En.Pvir[i]*=Punit/V; // /DIM removed
#  if PRESSURETENSOR&PT_KIN
    En.Ptens[i]=En.Pvir[i]+En.Pkin[i];
#  endif /*# PRESSURETENSOR&PT_KIN */
    }

  {
    static int pass=0;
    double x=SUM(En.Pvir)/DIM-En.vir*(Punit/DIM)/V; // /DIM added

    StaSet(0,lag.err,2,lag.n);
    StaAdd("Pvirerr [Pa]",x);

    if (!pass && fabs(x)>1000) {
      prt("WARNING: virial/3V=%g (%g Pa pressure-equivalent)\n\
         tr(Pvir)=%g (%g Pa)  dif=%g (%g Pa)\n\
         (more warnings suppressed)",
          En.vir*(1./DIM)/V,En.vir*(Punit/DIM)/V,
          SUM(En.Pvir)/Punit/DIM,SUM(En.Pvir)/DIM,x/Punit,x); // /DIM added 2x
      pass=1; }
  }
#  ifdef SLAB
    StaSet(0,lag.err,2,lag.n);
    StaAdd("Pvir xy-zz [Pa]",(En.Pvir[0]+En.Pvir[1]-2*En.Pvir[2])/DIM); // /DIM added
#    if PRESSURETENSOR&PT_KIN
    StaAdd("Pt [Pa]",((En.Pvir[0]+En.Pvir[1]-2*En.Pvir[2])
                    +(En.Pkin[0]+En.Pkin[1]-2*En.Pkin[2]))/DIM); // /DIM added
#    endif /*# PRESSURETENSOR&PT_KIN */
#  endif /*# SLAB */
    StaSet(0,2,2,lag.n);
#endif /*# PRESSURETENSOR&PT_VIR */

#if PRESSURETENSOR&PT_ANY
    loop (i,0,PT_DIM) {
      char *xx="xx\0yy\0zz\0yz\0zx\0xy"+i*3;

#  if PRESSURETENSOR&PT_VIR
      StaAdd(string("Pvir.%s [Pa]",xx),En.Pvir[i]);
#  endif /*# PRESSURETENSOR&PT_VIR */
#  if PRESSURETENSOR&PT_KIN
      StaAdd(string("Pkin.%s [Pa]",xx),En.Pkin[i]);
#  endif /*# PRESSURETENSOR&PT_KIN */
#  if PRESSURETENSOR&PT_MOL
      StaAdd(string("PKin.%s [Pa]",xx),En.PKin[i]);
#  endif /*# PRESSURETENSOR&PT_MOL */
#  if PRESSURETENSOR&PT_MOM
      if (i<3) {
        StaAdd(string("PKin2.%s [Pa^2]",xx),En.PKin2[i]*Sqr(Punit/V));
        StaAdd(string("PKin3.%s [Pa^3]",xx),En.PKin3[i]*Cub(Punit/V)); }
#  endif /*# PRESSURETENSOR&PT_MOM */
#  if (PRESSURETENSOR&PT_ANY) == PT_ANY
      StaAdd(string("Ptens.%s [Pa]",xx),En.Ptens[i]);
#  endif /*# (PRESSURETENSOR&PT_ANY) == PT_ANY */
    }
#endif /*# PRESSURETENSOR&PT_ANY */

  if (No.f_in) En.T_in=En.kin_in/No.f_in;
  if (No.f_tr>0) En.T_tr=En.kin_tr/No.f_tr;
} /* measureconstraints */

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                             constraint dynamics                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Formulae:
% ^^^^^^^^^
% M[a,b] = SUM{i} grad[i]ff[a].grad[i]ff[b] / m[i]
%
% P[a] = SUM{i} p[i].grad[i]ff[a] / m[i]         (Hamilton, obsolete)
%
% P[a] = SUM{i} F[i].grad[i]ff[a] / m[i]
%      + SUM{i,j} V[i].grad[i]grad[j]ff[a].V[j]  (Lagrange)
%
% where constraint ff[a] = (r[ia]-r[ja])**2/2 - BondLength**2/2
% ( ia = si[a].pair[0], ja = si[a].pair[1], BondLength**2 = si[a].bondq )
%
% F[i] = -grad[i]U is force, and  "." is scalar product
%
% Eq. for Lagrange multipliers: (note: g(Lagrange) = dg(Hamilton)/dt )
% ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
% SUM{b} M[a,b] g[b] = P[a]
%
% Eqs. of motion:
% ^^^^^^^^^^^^^^^
% m[i] dr[i]/dt = p[i] - SUM{a} g[a]grad[i]ff[a]
% dp[i]/dt      = F[i] + SUM{a}SUM{j}g[a]grad[i]grad[j].dr[i]/dt (Hamilton)
%
% m[i] ddr[i]/dt/dt = F[i] - SUM{a} g[a]grad[i]ff[a]             (Lagrange)
%
% Using matrix M to correct constraints: (see Lcorrect)
% ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
% Dr[i] = SUM{b} e[b] 1/m[i]*grad[i]ff[b]
% where e solves
% SUM{b} M[a,b] e[b] = -ff[a]
%
% Velocity constraints in the Lagrangian formalism:
% Dv[i] = SUM{b} ve[b] 1/m[i]*grad[i]ff[b]
% where ve solves
% SUM{b} M[a,b] ve[b] = -SUM{i}v[i]*grad[i]ff[a]
%
% Thermostats (Lagrange only):
% ^^^^^^^^^^^^^^^^^^^^^^^^^^^^
% Friction:
%   ddr[i]/dt/dt -= v[i] ln(Tkin/T)/2/tau.T
% Nose:
%   ddr[i]/dt/dt -= v[i] dlogs/dt
%   ddlogs/dt/dt = (Tkin/T-1)/tau.T^2
% Since ddr[i]/dt/dt is a `friction' force, the corresponding term appears
% also in P[]
%
% #define PRECOND :
%   uses pre-conditioned matrix M for conjugate gradients:
%     SUM{b} M*[a,b] g*[b] = P*[a]
%   where M*=diag.M.diag, P*=diag.P and the solution is g=diag.g*
%   ( . denotes matrix multiplication)
%   and diag=diag(M)^(-1/2)
%   It appears that about 2 iterations are spared but the control operations
%     cost about the same.
%   Not implemented for constraints corrections (option -c4)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#ifdef Lagrange                                        /* constraintdynamics */
/*** Lagrange equations of motion ***/
void constraintdynamics(ToIntPtr B, ToIntPtr A, ToIntPtr V)
#else /*# Lagrange */
/*** Hamilton equations of motion ***/
void constraintdynamics(ToIntPtr B, ToIntPtr A)
#endif /*#!Lagrange */
{
  int n,a,b,nc,ns,sp,i,j,k;
  static int irhs=0;
  vector *v,*f,*r,*ra=NULL,*rp=A->rp,*rpvel=NULL;
#ifdef PRECOND
  double *diag;
#endif /*# PRECOND */
#ifdef Hamilton
  vector *p,*dr,*dp,
#endif /*# Hamilton */
  molecule_t *mn;
  vector auxv;
  /* warning: even #ifdef FLOAT, the constraint calculations are in double */
  double *g=NULL,*aux,*P=NULL;
  double **gg=NULL;
  double z, Vlogs=0, Vlogse=0;
  M_t *M,*Msp;
  siteinfo_t *si;
  vector v_mol;
  double m_mol,mi;

  En.virc=0;

#ifdef ANCHOR
  if (option('k')==0) {
    static int pass=0;
    /* NOTE: 1st pass not reported because called to obtain initial forces */
    if (pass++)
      ERROR(("ANCHOR by constraints (-k0) not supported with Gear+Lagrange\n\
*** use SHAKE (-m2) instead")) }
#endif /*# ANCHOR */

  if (No.pred) allocarray(gg,No.pred);

  if (No.maxc) {
    rallocarray(ra,No.maxc); /* = r[i]-r[j] for constraint a=(i,j) */
    rallocarray(P,No.maxc);  /* rhs of constraint equation */
#ifdef PRECOND
    rallocarray(diag,No.maxc);
#endif /*# PRECOND */
    if (No.pred==0) rallocarrayzero(g,No.maxc); }

#ifdef Lagrange
  /* ! all kin.energies (En/kin, En.kin_tr, En.kin_in) are doubled here !*/
  switch (thermostat) {
    case T_NOSE: /* Nose */
      Vlogs=V->logs; break;
    case T_FRICTION: /* friction thermostat: must know En.kin=twice kin.en. in advance! */
      if (En.kin==0) Vlogs=0;
      else Vlogs=blogT(En.kin/No.f,T)/tau.T;
      break;
    case T_TR: /* friction thermostat - intermolecular only */
      Vlogse=blogT(En.kin_tr/No.f_tr,T)/tau.T;
      break;
    case T_IN: /* friction thermostat - intramolecular only */
      Vlogs=blogT(En.kin_in/No.f_in,T)/tau.T;
      break;
    case T_FRICTIONS: { /* decoupled friction thermostats, T_tr/T_in=T_tr_in */
      double T_tr=(No.f_in+No.f_tr)*T/(T_tr_in*No.f_in+No.f_tr);
      double T_in=(No.f_in+No.f_tr)*T/(No.f_in+No.f_tr/T_tr_in);

      Vlogs=blogT(En.kin_in/No.f_in,T_in)/tau.T;
      Vlogse=blogT(En.kin_tr/No.f_tr,T_tr)/tau.T; } }

  Vlogse-=Vlogs;
  rpvel=V->rp;
#else /*# Lagrange */
  gsq=0;
#endif /*#!Lagrange */

  /*** constraints: computing M and P ***/
  loop (n,FROM,No.N) { /* <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
    mn=cfg+n;
    sp=mn->sp;
    ns=mn->ns;
    si=spec[sp]->si;
    r=rof(mn,rp);
#ifdef Lagrange
    v=rof(mn,V->rp);
    f=rof(mn,B->rp);
#else /*# Lagrange */
    dr=rof(mn,B->rp);
    p=pof(mn,rp);
    dp=pof(mn,B->rp);
#endif /*#!Lagrange */

    if ((nc=mn->nc)) {

      if (No.pred) {
        loop (k,0,No.pred) gg[k]=gs[k]+mn->ig;
        g=gg[0]; }

      M=Msp=Ms[option('c')&1 ? n : sp];

      loop (a,0,nc) {
        i=si[a].pair[0]; j=si[a].pair[1];
        VVV(ra[a],=r[i],-r[j])

        M->Mab = (si[i].imass+si[j].imass)*SQR(ra[a]); /* a==b */
        while ( (b=(++M)->b) < a )
          M->Mab = M->imass*SCAL(ra[a],ra[b]);

#ifdef Lagrange
        VVV(auxv, =si[i].imass*f[i], -si[j].imass*f[j])
        P[a] = SCAL(auxv,ra[a]);
        VVV(auxv, =v[i], -v[j])
        P[a] += SQR(auxv);
        if (tau.T) P[a] -= 2*Vlogs*SCAL(auxv,ra[a]);
#else /*# Lagrange */
        VVV(auxv, =si[i].imass*p[i], -si[j].imass*p[j])
        P[a] = SCAL(auxv,ra[a]);
#endif /*#!Lagrange */

        /*** predictor of Lagrange multipliers g ***/
        if (No.pred>0) {
          z=0;
          loop (i,0,No.pred) z+=binom[i]*gg[i][a];
          gg[No.pred-1][a]=z; }
        } /* a */

      if (No.pred) g=gg[No.pred-1];
      else memset(g,0,nc*sizeof(g[0]));

#ifdef PRECOND
      /* find the diagonal */
      if (option('c')&1) Error("-c&1 not allowed with PRECOND");
      /* ...should pass diag to Lcorrect (norm) and use it in the same way */
      M=Msp;
      loop (a,0,nc) {
        diag[a]=1./sqrt(M->Mab); /* a==b */
        while ( (b=(++M)->b) < a ); }

      /* calculating diag.M.diag and diag.P */
      M=Msp;
      loop (a,0,nc) {
        M->Mab=1; /* a==b */
        P[a]*=diag[a];
        g[a]/=diag[a]; /* to fix the predictor of g */
        while ( (b=(++M)->b) < a ) M->Mab*=diag[a]*diag[b]; }
#endif /*# PRECOND */

      /*** solving the eqs. for Lagrange multipliers ***/
      constrit[sp].nit[LAGR_MULT] += omega==0
        ? conjgrad  (nc,g,Mvec,Msp,P, irhs<No.pred && eps>=1?1e-12:eps)
        : directiter(nc,g,Msp,P,omega,irhs<No.pred && eps>=1?1e-12:eps);

#ifdef PRECOND
      loop (a,0,nc) g[a]*=diag[a];
#endif /*# PRECOND */

      } /* if (nc) */

    /*** equations of motion ***/
#ifdef Lagrange
    loop (a,0,nc) {
      i=si[a].pair[0]; j=si[a].pair[1];

      /*** forces equiv to constraint motions, virial ***/
      VV(auxv, =g[a]*ra[a])
      VV(f[i], -=auxv)
      VV(f[j], +=auxv)
      En.virc -= g[a]*si[a].bondq;
#  if PRESSURETENSOR&PT_VIR
      En.Pvir[0] -= g[a]*ra[a][0]*ra[a][0];
      En.Pvir[1] -= g[a]*ra[a][1]*ra[a][1];
      En.Pvir[2] -= g[a]*ra[a][2]*ra[a][2];
#    if PRESSURETENSOR&PT_OFF
      En.Pvir[3] -= g[a]*ra[a][1]*ra[a][2];
      En.Pvir[4] -= g[a]*ra[a][0]*ra[a][2];
      En.Pvir[5] -= g[a]*ra[a][0]*ra[a][1];
#    endif /*# PRESSURETENSOR&PT_OFF */
#  endif /*# PRESSURETENSOR&PT_VIR */
      } /*a*/

    /*** thermostats & acceleration from forces (kin. energy now in rhs) ***/

    /*** special for decoupled inter/intra thermostat ***/
    if (thermostat>=T_TR) {
      VO(v_mol,=0)
      m_mol=0;
      loop (i,0,ns) {
        mi=si[i].mass;
        m_mol+=mi;
        VV(v_mol,+=mi*v[i]) }

      VO(v_mol,/=m_mol) /* velocity of CM of a molecule */

      loop (i,0,ns)
        VVVV(f[i], =si[i].imass*f[i], -Vlogs*v[i], -Vlogse*v_mol) }

    /*** Nose or friction thermostat (not decoupled) - Vlogs see above ***/
    else if (tau.T)
      loop (i,0,ns)
        VVV(f[i], =si[i].imass*f[i], -Vlogs*v[i])

    /*** no thermostat ***/
    else
      loop (i,0,ns)
        VO(f[i], *=si[i].imass)

#else /*** Hamilton (obsolete, no pressure tensor) ***/ /*# Lagrange */

    loop (i,0,ns) VV(dr[i], =p[i])

    loop (a,0,nc) {
      i=si[a].pair[0]; j=si[a].pair[1];

      gsq += g[i]*g[i];

      VV(auxv, =g[a]*ra[a])
      VV(dr[i], -=auxv)
      VV(dr[j], +=auxv)
      } /* a */

    loop (i,0,ns) VO(dr[i], *=si[i].imass)

    /*** dp[i]+=g[a]*(dr[i]-dr[j]), dp[j]-=dtto ***/
    loop (a,0,nc) {
      i=si[a].pair[0]; j=si[a].pair[1];
      loop (b,0,3) {
        z=g[a]*(dr[i][b]-dr[j][b]);
        dp[i][b]+=z; dp[j][b]-=z; } }

    /*** kinetic energy ***/
    if (measure) {
      loop (i,0,ns) En.kin += SQR(p[i])*si[i].imass;
      loop (a,0,nc) En.kin -= g[a]*P[a]; }
#endif /*#!Lagrange */

#if 0
/* very old - option -c changed */
    if (option('c')==2 && nc)
      constrit[sp].nit[COR_BOTH] += Scorrect(mn,rof(mn,rp),rpvel,omegac,epsc);
#endif /*# 0 */
    } /* n >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> */


  if (No.pred) {
    /*** dirty trick: pointer shuffling ***/
    k=No.pred-1; aux=gs[k];
    while (k) { gs[k]=gs[k-1]; k--; }
    gs[0]=aux;
    free(gg); }

  if (No.maxc) release(ra);

  /* final energies and statistics: repeated in Shake */

  En.T=En.kin/No.f;
  /* ! for historical reasons, En.kin_tr and En.kin_in are still doubled! */
  En.kin/=2;
#ifdef POLAR
  En.pot += En.self;
#endif /*# POLAR */
  En.pot+=En.el;
  En.tot=En.kin+En.pot;

  /*.....if (measure) fprintf(stderr,"%12.2f %12.2f %12.2f %12.2f %12.2f %12.2f\n", En.tot,En.kin,En.pot,En.el,En.intra,En.self);*/

  if (thermostat==T_NOSE) {
    En.tot += No.f*T*(A->logs+Sqr(tau.T)/2*Sqr(Vlogs));
    B->logs = (En.T/T-1)/Sqr(tau.T); }

#ifdef PRTKIN
  prt_("T_in = %.2f",En.kin_in/No.f_in);
  prt("   T_tr = %.2f",En.kin_tr/No.f_tr);
#endif /*# PRTKIN */

  if (measure==1) measureconstraints();

#if 0
/* very old - option -c changed */
  if (option('c')<3) {
#  ifdef Lagrange
    En.r2=constrainterror(A,V);
    En.v2=vconstrainterror;
#  else /*# Lagrange */
    En.r2=constrainterror(A,NULL);
#  endif /*#!Lagrange */
    }
#endif /*# 0 */

  irhs++;

#ifdef COOK
  if (FROM) /* to keep 1st FROM molecules on place */
    loop (n,0,FROM) {
      mn=cfg+n;
      ns=mn->ns;
      memset(rof(mn,V->rp),0,ns*sizeof(vector));
      memset(rof(mn,B->rp),0,ns*sizeof(vector)); }
#endif /*# COOK */

  CPUtime("constr");

#ifdef POLAR
  if (option('p')/10%10==0) {
    if (init>=0 && init<3)
      ERROR(("Gear + Car-Parrinello NOT SUPPORTED"))
    else
      WARNING(("Gear + Car-Parrinello: BRUTE FORCE INITIALIZATION"))

    loop (n,0/*FROM ?*/,No.N) {
      mn=cfg+n;
      sp=mn->sp;
      ns=mn->ns;
      si=spec[sp]->si;
      r=polarrof(mn,rp);
#  ifdef Lagrange
      v=polarrof(mn,V->rp);
      f=polarrof(mn,B->rp);
#  else /*# Lagrange */
#    error Hamilton & POLAR
#  endif /*#!Lagrange */
      loop (i,0,ns) VVVO(r[i],=v[i],=f[i],=0); } }
#endif /*# POLAR */

} /* constraintdynamics */


/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                   SHAKE                                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#ifdef Lagrange

#  ifdef SHAKE

#    ifdef ANCHOR
#      include "jacobi.c"
#    endif /*# ANCHOR */

void Shake(double eps, double prob) /* =============================== Shake */
/*
  Verlet integration (equivalent to leap-frog)
  with SHAKE constraint dynamics for bonds and similar external constraints
  eps = accuracy (if eps<1)
      = # of Shake iteration sweeps over all sites (if eps>=1, |SHAKE|=1 only)
  omega = relaxation parameter (omega=1 is direct iteration)
  data:
    rof(molecule *,a[0]->rp) = r(t)
    rof(molecule *,a[1]->rp) = r(t)-r(t-h) = h*v(t-h/2)
    rof(molecule *,a[2]->rp) = forces; used to keep r(t+h) during shaking

  POLAR:
    polarrof(molecule *,a[0]->rp) = dr(t) (rel.aux.site of ind. dipole)
    polarrof(molecule *,a[1]->rp) = dr(t-h)
    polarrof(molecule *,a[2]->rp) = used in forces calculation

  Compile-time versions:
    SHAKE>0: simplified formula, equivalent upto h^2,
             usu more efficient for complex systems
    SHAKE<0: `exact' formula using the scalar product,
             perhaps better for diatomics
    |SHAKE|=1: update in sweeps (generally better)
    |SHAKE|=2: update for unprecise bonds only
               (better for e.g. mixture of complex molecule + diatomics)

  History:
    1992 first version
    1995 revisited
    2001 Maxwell/Andersen thermostat
    2003 ANCHOR constraints, |SHAKE|=1 only
    2008 pressure tensor
*/
{
  vector rij,pij;
  int al,n,i,j,sp,nc,ns=-1,it,maxit0=-16*log(eps*0.1);
  molecule_t *mn;
  vector *r,*p,*rp=a[0]->rp,*r1,*rp2=a[2]->rp,*v;
  siteinfo_t *si;
  double rxp,bondq,z,ze,mi,mj,sq;
  vector v_mol;
  vector v_mol1,v_mol2;
  double m_mol;
  double hh=h*h,Thh=T*hh,hhh=hh/2,hscale;
  double omega; /* = omegac/2 */
  int CMbasedthermostat=thermostat==T_ANDERSEN_CM || thermostat==T_MAXWELL_CM;

  /* Nose-Hoover; arrays are allocated forever */
  static double **Vhist; /* Vhist[0]=X(t)-X(t-h) (pointer passed); X=logs,r
                            Vhist[1]=X(t-h)-X(t-2h) ... */
  static int k; /* predictor extra length, actual value, growing up to No.pred */
  static double *b; /* coefficients */
  static double *Vpred; /* Vpred[0]=h*d X(t)/dt, predicted */
  static vector *vpred; /* vpred=pointer to predicted h*v[0] */
  static double ENose=0; /* 0 unless Nose */

#    if SHAKE==1 || SHAKE==-1
  int ifit=eps>=1;
  int maxit = (int)eps;
  double maxsq;
#    else /*# SHAKE==1 || SHAKE==-1 */
  int *moved=NULL,*moving,done;
#    endif /*#!SHAKE==1 || SHAKE==-1 */

#    ifdef ANCHOR
  vector cf={0,0,0};
  int notinplace=0;
#    endif /*# ANCHOR */

  t+=h;
  En.virc=0;

#    ifndef FREEBC
  VV(box.Lh,=0.5*box.L)
#    endif /*# FREEBC */

  box.cq=Sqr(box.cutoff);
  En.kin=En.kin_tr=0;

  if (option('c')&2) {
    En.r1=constrainterror(a[0],a[1]);
    En.v1=vconstrainterror/h; }

  depend_rv(a[0],a[1]);

#    ifdef POLAR
  /* pred1.c is old version with -@%=A[0] -_%=A[1]: WARNING - options changed */
#      include "pred2.c"

  if (option('p')/10%10==0)
    mechpolar(a[2],a[0]);
  else do { /* iterate self-field */
#    endif /*# POLAR */

    zeroEn();
    forces(a[2],a[0]);

#    ifdef POLAR
    } while (!selffield(a[2],a[0],epsp));
#    endif /*# POLAR */

#    if PRESSURETENSOR&PT_VIR
  if (measure) memset(PTvirc,0,sizeof(PTvirc));
#    endif /*# PRESSURETENSOR&PT_VIR */

  depend_f(a[0],a[2]);
  /*.....if (measure) dipolemoments(a[0]);*/

#    ifdef SHEAR
  Shear(a[2],a[0],a[1],h);
#    endif /*# SHEAR */

  /*** independent pressure measurement via <dU/dV>
       and errors in induced dipoles: pass 1 */
  if (measure) measureP(1);

#    if VERLET==4
  if (thermostat!=T_NOSE) ERROR(("implementation limitation\n\
*** version VERLET==4 (predicted velocities) requires a Nose-Hoover thermostat"))
#    endif /*# VERLET==4 */

  if (thermostat==T_NOSE) {
    static int nose_k=-1,lastk=-1;

    if (nose_k>=0 && No.pred!=nose_k)
      ERROR(("SHAKE+Nose velocity predictor length unexpectedly changed (%d->%d)",nose_k,No.pred))

    if (nose_k<0 && No.pred==0)
      WARNING(("SHAKE+Nose without predictor (option -p0) is imprecise \n\
*** (Hamiltonian drift ~O(h^1))"))
    nose_k=No.pred;

    if (!Vpred) {
      allocarray(b,No.pred+1); /* predictor coefficients */
      allocarray(Vpred,No.eq); /* predicted value */
      vpred=(vector*)(Vpred+1);
      allocarray(Vhist,No.pred+1); /* the history */
      loopto (j,1,No.pred) allocarray(Vhist[j],No.eq);
      k=0; /* no history known now */ }

    Vhist[0]=&a[1]->logs; /* pointer passed - not allocated */

    /* generate predictor coefficients */
    if (k!=lastk) {
      b[0]=(double)(2*k+1)/(k+1);
      loop (j,0,k) b[j+1]=-b[j]*(double)(k-j)/(k+j+2);
      prt("velocity predictor for Verlet+SHAKE with Nose-Hoover, k=%d (max %d):",k,No.pred);
      loopto (j,0,k) prt("  b[%d]=%.15g",j,b[j]); }
    lastk=k;


    /* the predictor */
    loop (i,0,No.eq) {
      Vpred[i]=0;
      loopto (j,0,k) Vpred[i]+=b[j]*Vhist[j][i]; }

    /* increase k up to No.pred */
    k=min(No.pred,k+1);

    /* history shift */
    for (j=k; j; j--)
      loop (i,0,No.eq) Vhist[j][i]=Vhist[j-1][i]; }

  /* <<<<< loop over all molecules w. Verlet+SHAKE starts here <<<<< */

  loop (n,FROM,No.N) {
    mn=cfg+n;
    sp=mn->sp;
    r=rof(mn,rp);
    p=rof(mn,rp2);
    v=rof(mn,vpred);
    r1=rof(mn,a[1]->rp);
    si=spec[sp]->si;
    nc=mn->nc;
    omega=constrit[sp].omega; /* HALF omegac, the relaxation parameter */

#    if SHAKE>0 /* simplified and in general more efficient algorithm */
#      define CALCULATE_RXP rxp=bondq;
#    else /* more complicated algorithm using angle between p & r */ /*# SHAKE>0 */
#      define CALCULATE_RXP rxp=SCAL(rij,pij) \
         if (rxp<bondq*1e-6) ERROR(( \
           "constraint failure\n" \
           "it=%d %d[%d-%d]:(%.3f %.3f %.3f)->(%.3f %.3f %.3f)", \
           it,n,i,j,pij[0],pij[1],pij[2],rij[0],rij[1],rij[2]));
#    endif /*#!SHAKE>0 */

#    if SHAKE==1 || SHAKE==-1 /* constraints updated in sweeps */
    ns=mn->ns;

#      ifdef ANCHOR
#        define FORCEUNIT (massunit*lengthunit/Sqr(timeunit))
#        define TORQUEUNIT (massunit*Sqr(lengthunit/timeunit))
#        define PRINTANCHOR(U,V) if (anchor.f) fprintf(anchor.f," %12g %12g %12g ", \
  (U)*V[0],(U)*V[1],(U)*V[2])
   /* measure force on center-of-mass and torque
      set zeros to cf of groups
      (independent of anchoring): */
#        include "anchorm.c"
#      endif /*# ANCHOR */

/* Verlet step (for 1 molecule) + possible Andersen/Maxwell thermostat */
#      include "verlet.c"

    if (!ifit) maxit=maxit0*(nc+4);

    it=0;
    do { it++;
      maxsq=0;

#      ifdef ANCHOR
     /* anchor site: */
#        include "anchors.c"
     /* anchor CM of group(s): */
#        include "anchorg.c"
#      endif /*# ANCHOR */

      loop (al,0,nc) {
        i=si[al].pair[0]; j=si[al].pair[1];

        sq=bondq=si[al].bondq;
        VVV(pij,=p[i],-p[j])
        sq -= SQR(pij);
        Max(maxsq,fabs(sq/bondq))

        VVV(rij,=r[i],-r[j])
        CALCULATE_RXP
        mi=si[i].imass; mj=si[j].imass;
        sq *= omega/(mi+mj);
#      include "virc.c"
        sq /= rxp;
        VO(rij,*=sq)
        VV(p[i],+=mi*rij)
        VV(p[j],-=mj*rij) } /*al*/

      } while ((ifit || maxsq>eps) && it<maxit);

#      ifdef ANCHOR
    /* constrain CM, axes, and print site/cm constraint force */
#        include "anchora.c"
#      endif /*# ANCHOR */

   if (!ifit && it==maxit) {
     int ii;
     ERROR(("Shake: species=%d molecule=%d: too many iterations\n\
*** (selecting (i)gnore in interactive run will dump the molecule)",mn->sp,n))
     loop (ii,0,ns) prt("%d: %g %g %g",ii,VARG(r[ii])); }

#    else /*# SHAKE==1 || SHAKE==-1 */
/* SHAKE=+-2: info about moved sites is kept, eps<1 only */
#      ifdef ANCHOR
#        error ANCHOR not supported with this version of SHAKE
#      endif /*# ANCHOR */

    if (ns<mn->ns) {
      ns=mn->ns;
      if (moved!=NULL) { free(moved); free(moving); }
      allocarray(moved,ns);
      allocarray(moving,ns); }

/* Verlet step (for 1 molecule) + possible Andersen/Maxwell thermostat */
#      include "verlet.c"

    loop (i,0,ns) { moving[i]=0; moved[i]=1; }

    done=it=0;
    do {
      done=1;

      if (it>maxit0*(nc+4)) {
        int ii;
        ERROR(("Shake: species=%d molecule=%d: too many iterations\n\
*** (selecting (i)gnore in interactive run will dump the molecule)",mn->sp,n))
        loop (ii,0,ns) prt("%d: %g %g %g",ii,VARG(r[ii])); }

      loop (al,0,nc) { it++;
        i=si[al].pair[0]; j=si[al].pair[1];

        if (moved[i] || moved[j]) {

          sq=bondq=si[al].bondq;
          VVV(pij,=p[i],-p[j])
          sq -= SQR(pij);

          if (fabs(sq) > eps*bondq) {

            VVV(rij,=r[i],-r[j])
            CALCULATE_RXP
            mi=si[i].imass; mj=si[j].imass;
            sq *= omega/(mi+mj);
#      include "virc.c"
            sq /= rxp;
            VO(rij,*=sq)
            VV(p[i],+=mi*rij)
            VV(p[j],-=mj*rij)

            moving[i]=moving[j]=1; done=0; }
          } /* i and j moved */
        } /*a*/
      loop (i,0,ns) { moved[i]=moving[i]; moving[i]=0; }
      } while (!done);

/*.....if (n==0) put(it)*/

#    endif /*#!SHAKE==1 || SHAKE==-1 */

/*
  kinetic energy, pressure tensor, and final a[0], a[1] update
  four versions using different formulas for velocity
*/
#    define V1 r1[i] /* [r(t)-r(t-h)]/h */
#    if VERLET==0
#      include "shakev0.c"
#    elif VERLET==1
#      include "shakev1.c"
#    elif VERLET==2
#      include "shakev2.c"
#    elif VERLET==3
#      include "shakev3.c"
#    elif VERLET==4
#      include "shakev4.c"
#    elif VERLET==30
#      include "shakev30.c"
#    else /*? VERLET==30 */ /*#!VERLET==0!VERLET==1!VERLET==2!VERLET==3!VERLET==4!VERLET==30 */
#      error VERLET should be #defined as one of 0,1,2,3,4,5
#    endif /*? VERLET==30 */ /*#!VERLET==0!VERLET==1!VERLET==2!VERLET==3!VERLET==4!VERLET==30 */
#    undef V1

    if (nc) constrit[sp].nit[COR_BONDS] += it;
    } /*n*/
  /* >>>>>* loop over all molecules w. Verlet+SHAKE ends here >>>>> */

#    ifdef POLAR
#      include "shakecp.c"
#    endif /*# POLAR */

#    if SHAKE==2 || SHAKE==-2
  if (moved!=NULL) { free(moved); free(moving); }
#    endif /*# SHAKE==2 || SHAKE==-2 */

  /* TWICE the respective kinetic energies: */
#    if VERLET==3 || VERLET==30
  hscale = 0.5/hh;
#    elif VERLET==1
  hscale = 0.25/hh;
#    else /*#!VERLET==3 || VERLET==30!VERLET==1 */
  hscale = 1./hh;
#    endif /*#!VERLET==3 || VERLET==30!VERLET==1 */

  En.kin *= hscale;
  En.kin_tr *= hscale;

#    if VERLET==2
#      define HSCALE (i<DIM?hscale:hscale/2)
#    else /*# VERLET==2 */
#      define HSCALE hscale
#    endif /*#!VERLET==2 */

#    if PRESSURETENSOR&PT_KIN
  loop (i,0,PT_DIM) En.Pkin[i]*=HSCALE;

#      if VERLET==30
  hscale*=4;
#      endif /*# VERLET==30 */

#      if PRESSURETENSOR&PT_MOL
  loop (i,0,PT_DIM) En.PKin[i]*=HSCALE;
#        if PRESSURETENSOR&PT_MOM
  loop (i,0,DIM) {
    En.PKin2[i]*=Sqr(hscale);
    En.PKin3[i]*=Cub(hscale); }
#        endif /*# PRESSURETENSOR&PT_MOM */
#      endif /*# PRESSURETENSOR&PT_MOL */
#    endif /*# PRESSURETENSOR&PT_KIN */

  En.kin_in=En.kin-En.kin_tr;

#    ifdef SHEAR
  /*
    systematic flow kinetic energy removed
    NOTE: in some cases, the rest of the kinetic energy is calculated later
  */
  En.kin-=En.kinshear;
  En.kin_tr-=En.kinshear;
#    endif /*# SHEAR */

#    ifdef PRTKIN
  prt_("T_in = %.2f",En.kin_in/No.f_in);
  prt("   T_tr = %.2f",En.kin_tr/No.f_tr);
#    endif /*# PRTKIN */

  En.T=En.kin/No.f; /* NB: En.kin still doubled here ! */

  if (measure) measureP(2);

/* decoupled friction thermostats */
  if (thermostat>=T_TR) {
    z=ze=0;

    switch (thermostat) {
      case T_TR: /* friction thermostat - intermolecular only */
        ze=blogT(En.kin_tr/No.f_tr,T)*(h/tau.T);
        break;
      case T_IN:
        z=blogT(En.kin_in/No.f_in,T)*(h/tau.T);
        break;
      case T_FRICTIONS: {
        double T_tr=(No.f_in+No.f_tr)*T/(T_tr_in*No.f_in+No.f_tr);
        double T_in=(No.f_in+No.f_tr)*T/(No.f_in+No.f_tr/T_tr_in);

        z=blogT(En.kin_in/No.f_in,T_in)*(h/tau.T);
        ze=blogT(En.kin_tr/No.f_tr,T_tr)*(h/tau.T); } }

    ze-=z;
    loop (n,FROM,No.N) {
      si=spec[cfg[n].sp]->si;
      r1=rof(&cfg[n],a[1]->rp);
      ns=cfg[n].ns;
      VO(v_mol,=0)
      m_mol=0;
      loop (i,0,ns) {
        mi=si[i].mass;
        m_mol+=mi;
        VV(v_mol,+=mi*r1[i]) }

      VO(v_mol,/=m_mol) /* h*velocity of CM of a molecule */

      loop (i,0,ns) VVV(r1[i], -=z*r1[i], +ze*v_mol) } }

  else if (thermostat==T_FRICTION) {
    /* friction thermostat */
/*.....    z=pow(En.T/T,-h/2/tau.T);*/
    z=exp(blogT(En.T,T)*-h/tau.T);
    loop (n,FROM,No.N) {
      r1=rof(&cfg[n],a[1]->rp);
      loop (i,0,cfg[n].ns) VO(r1[i],*=z) } }

  else if (thermostat==T_NOSE) {
    real llogs=a[0]->logs,ldlogs=a[1]->logs;

    /* Verlet (=leap-frog here) for logs */
    ENose = a[0]->logs;
    a[1]->logs+=hh*(En.T/T-1)/Sqr(tau.T);
    a[0]->logs+=a[1]->logs; /* only good for the total Hamiltonian */

#    if VERLET==0
    ENose += Sqr(tau.T*ldlogs/h)/2;
#    elif VERLET==1
    ENose += Sqr(tau.T*(ldlogs+a[1]->logs)/h)/8;
#    elif VERLET==2
    ENose += Sqr(tau.T/h)*ldlogs*a[1]->logs/2;
#    elif VERLET==3 || VERLET==30
    ENose += Sqr(tau.T/h)*(Sqr(ldlogs)+Sqr(a[1]->logs))/4;
#    elif VERLET==4
    ENose += Sqr(tau.T*Vpred[0]/h)/2;
#    else /*#!VERLET==0!VERLET==1!VERLET==2!VERLET==3 || VERLET==30!VERLET==4 */
#      error VERLET should be #defined as one of 0,1,2,3,4,5
#    endif /*#!VERLET==0!VERLET==1!VERLET==2!VERLET==3 || VERLET==30!VERLET==4 */
    ENose*=No.f*T; }

  En.kin/=2;

#    ifdef POLAR
  En.pot += En.self;
#    endif /*# POLAR */
  En.pot+=En.el;

  En.tot=En.kin+En.pot+ENose;

  //  if (measure) prt("%.9g %.9g %.9g %.9g NOSE",En.kin,En.pot,No.f*T*a[0]->logs,No.f*T*Sqr(tau.T*Vpred[0]/h)/2);

  if (measure) if ( !(En.tot==0) && !(En.tot>0) && !(En.tot<0) )
    ERROR(("numeric problem: En.kin=%g En.pot=%g",En.kin,En.pot))

  En.virc/=hh; /* note: factor 1/2 is incl. in omega */
#    if PRESSURETENSOR&PT_VIR
  loop (n,0,PT_DIM) PTvirc[n]/=hh;
#    endif /*# PRESSURETENSOR&PT_VIR */

  if (measure==1) measureconstraints();

#    ifdef ANCHOR
  if (anchor.f) {
    if (notinplace) {
      fprintf(anchor.f," not in place or too large force/torque");
      fprintf(stderr,"ANCHOR: not in place or too large force/torque (%d times)\n",notinplace); }
    fprintf(anchor.f," %g\n",t); }
#    endif /*# ANCHOR */

  CPUtime("Shake");
} /* Shake */

#  endif /*# SHAKE */

#endif /* Lagrange */ /*# Lagrange */

#ifdef POLAR
int selffield(ToIntPtr B,ToIntPtr A,double epsp) /**************** selffield */
/*
  self-field: one iteration -- returns 1 if sufficiently accurate
*/
{
  molecule_t *mn;
  siteinfo_t *si;
  int ns,sp,n,i,nmax=-1,imax=-1;
  vector *fc,*fpol,*rpol,dr,dd;
#  if POLAR&8
  vector *rc;
#  endif /*# POLAR&8 */
  int ierr=0;
  double rr,f,maxrrold=0,maxrr=0,maxerr=0,err=0,iomegap,fdr;
#  ifndef FREEBC
  double byerd;
#  endif /*# FREEBC */

  static int option_v=-3333;
  static double oldt=-3e33,maxerrold;
  static int nit,nbad;

  /* maxerr|err is max|stdev error of induced dipolemoment */

#  define LAMBDA -1
 /*
   #ifdef LAMBDA: SPECIAL for finding eigenvalues
 LAMBDA>=0:
   if epsp<=0 then 25 files xxx.a, xxx.b, ..., xxx.y are generated with
   positions of auxiliary charges to be later read by program lambda.c to find
   eigenvalues of the convergence matrix M
   the value of LAMBDA denotes site in 1st 25 molecules recorded (x-coord)
   in addition, zzz.z contains max error
   epsp<0 means initial perturbation of used 25 dipoles by epsp
   the value LAMBDA refers to site number in a molecule: it is your
     responsibility to set it correctly!
 LAMBDA<0:
   does not generate files, does not stop unless ^C, sig=2
   epsp<0 means initial random perturbation of all dipoles amplitude
   sig=1 multiplies all dipoles by 10
   sig=5 multiplies all dipoles by 0.1
 */

#  ifdef LAMBDA
  static FILE *xxx[26];
  static ToIntPtr twofirst[2];
  static int pass;
#    if POLAR&8
  int tozz;
#    endif /*# POLAR&8 */


  if (epsp<=0) {
    if (!pass) {
      option_v=8;
#    if LAMBDA>=0
      loop (i,0,26) {
        static char xxxnm[]="xxx.?";

        xxxnm[4]=i+'a';
        if ( (xxx[i]=fopen(xxxnm,"wt"))==NULL ) ERROR(("cannot open %s",xxxnm))
        /* perturbed: for more accurate finding eigenvalues */
        polarrof(cfg+i,A->rp)[LAMBDA][0]+=epsp; }
      prt("^C and selecting 2 interrupts calculations");
#    else /* LAMBDA<0 */ /*# LAMBDA>=0 */
      prt("^C changes meaning. Select:\n\
  1 to multiply all dipoles 10x\n\
  5 to multiply all dipoles 0.1x\n\
  2 to stop");
      loop (n,FROM,No.N) {
        mn=cfg+n;
        ns=mn->ns;
        sp=mn->sp;
        si=spec[sp]->si;
        rpol=polarrof(mn,A->rp);
        loop (i,0,ns) if (si[i].alpha_qq) VO(rpol[i],+=rndgauss()*epsp) }
#    endif /*#!LAMBDA>=0 */
    }
    if (pass<2) {
      sdsalloc(twofirst[pass],sizeof(ToInt)-sizeof(vector)+No.s*2*sizeof(vector));
      sdscopy(twofirst[pass],A);
      pass++; }
  }
#  endif /*# LAMBDA */

  iomegap=1-omegap;
  En.self=0;
  if (option_v==-3333) option_v=option('v'); /* |=8 if bad convergence */

  loop (n,FROM,No.N) {
    mn=cfg+n;
    ns=mn->ns;
    sp=mn->sp;
    si=spec[sp]->si;
#  if POLAR&8
    rc=rof(mn,A->rp);
#  endif /*# POLAR&8 */
    rpol=polarrof(mn,A->rp);
    fc=rof(mn,B->rp);
    fpol=polarrof(mn,B->rp);
    loop (i,0,ns) {
      VVV(dd,=dr,=rpol[i])
      rr=SQR(dr);
      Max(maxrrold,rr)
      if (rr>=Sqr(el.minqq))
        ERROR(("POLARold: %d[%d] aux charge dist = %g",n,i,sqrt(rr)))

      /* NOTE (9/2001): the virial of force equals -2*self energy and
         is treated separately because of the Ewald corrections */

#  if defined(COULOMB) && COULOMB<0
      /* Ewald correction for dipole self-term */
      f=si[i].charge*si[i].chargepol;
      En.el+=f*exacterud_sqrt(rr,&byerd);
      f*=byerd;
// VVO(fc[i],-=dr,*=f) VV(fpol[i],+=dr) /* note sign: dr=pol-center */
#    define ENPVIR En.Pvir
      MADDFORCES(fpol[i],fc[i],f,dr)
#    undef ENPVIR
#  endif /*# defined(COULOMB) && COULOMB<0 */

#  if POLAR&8
      /* axial polarizability */
      if ( (tozz=si[i].tozz)>=0 ) {
#    if POLAR&2
        if (si[i].Esat)
          /* saturation */
          ERROR(("axial saturated polarizability not supported"))
        else
#    else /*# POLAR&2 */
        {
          /* axial+no saturation */
          vector drtozz,Ezz,Exx,rpolnew;
          double zz,rr,rf,a;

          VVV(drtozz,=rc[tozz],-rc[i])
          rr=SQR(drtozz); rf=SCAL(fpol[i],drtozz);
          zz=rf/rr;
          VV(Ezz,=zz*drtozz)
          VVV(Exx,=fpol[i],-Ezz)
          VVV(rpolnew,=si[i].alpha_qq*Exx,+si[i].alphazz_qq*Ezz)

          if (iomegap) /* relaxation */
            VVVV(dd,-=rpol[i],=iomegap*rpol[i],+omegap*rpolnew)
          else /* omegap=1: no relaxation */
            VVV(dd,-=rpol[i],=rpolnew)

          /* forces passed to both atoms */
          /* WARNING: not for negative -p */
          VV(fc[i],+=fpol[i])
          VVV(drtozz,=fpol[i],-zz*drtozz)
          a=rf/rr*(si[i].alpha_qq-si[i].alphazz_qq);
          VO(drtozz,*=a)
          VV(fc[i],+=drtozz)
          VV(fc[tozz],-=drtozz) }
#    endif /*#!POLAR&2 */
        /* forces passed to both atoms -- eventually move here after axial
           saturated is implemented */
      }
      else
#  endif /*# POLAR&8 */
        {
#  if POLAR&2
          /* new pos of auxiliary charges = new dipole moments */
          if (si[i].Esat) {
            /* saturation */
            double asat=si[i].alpha_qq
              / (0.5+sqrt(0.25+si[i].alpha_qq*SQR(fpol[i])/si[i].Esat));
            if (iomegap) /* relaxation */
            VVVV(dd,-=rpol[i],=iomegap*rpol[i],+(omegap*asat)*fpol[i])
          else /* omegap=1: no relaxation */
            VVV(dd,-=rpol[i],=asat*fpol[i]) }
        else
#  endif /*# POLAR&2 */
          {
            /* standard: no saturation */
            if (iomegap) /* relaxation */
              VVVV(dd,-=rpol[i],=iomegap*rpol[i],+(omegap*si[i].alpha_qq)*fpol[i])
            else /* omegap=1: no relaxation */
              VVV(dd,-=rpol[i],=si[i].alpha_qq*fpol[i])
          }
        /* forces passed to the central atom */
        VV(fc[i],+=fpol[i]) }

      /* twice the dipole self-energy */
      f=SCAL(rpol[i],fpol[i]);
#  if POLAR&2
      if (si[i].Esat)
        En.self += si[i].Esat*log1(f/si[i].Esat);
#    if PRESSURETENSOR&PT_VIR
#      error POLAR&2 & PRESSURETENSOR&PT_VIR - no support
#    endif /*# PRESSURETENSOR&PT_VIR */
      else
#  endif /*# POLAR&2 */
        En.self += f; /* = -virial (will be added to En.vir later) */

#  if PRESSURETENSOR&PT_VIR
      /* the sign is not clear to me, but this works... */
      VVV(En.Pvir,-=rpol[i],*fpol[i])
#    if PRESSURETENSOR&PT_OFF
#      if 1
      En.Pvir[3]-=rpol[i][1]*fpol[i][2];
      En.Pvir[4]-=rpol[i][2]*fpol[i][0];
      En.Pvir[5]-=rpol[i][0]*fpol[i][1];
#      else /*# 1 */
      /* debug: should be the same */
      En.Pvir[3]-=fpol[i][1]*rpol[i][2];
      En.Pvir[4]-=fpol[i][2]*rpol[i][0];
      En.Pvir[5]-=fpol[i][0]*rpol[i][1];
#      endif /*#!1 */
#    endif /*# PRESSURETENSOR&PT_OFF */
#  endif /*# PRESSURETENSOR&PT_VIR */

#  if 0
      /* old version: equivalent if SCF exactly solved */
      En.self += si[i].alpha_qq*SQR(fpol[i]);
#  endif /*# 0 */

#  if 0 /* MOVED and updated for axial... */
      /* forces passed to the central atom */
      VV(fc[i],+=fpol[i])
#  endif /*# 0 */

      /* check that aux charges are close enough to the center */

      rr=SQR(rpol[i]);
      if (rr>=Sqr(el.minqq))
        ERROR(("POLARnew: %d[%d] aux charge dist = %g",n,i,sqrt(rr)))

      if (rr>maxrr) { maxrr=rr; imax=i; nmax=n; }

/*.....if (rr>1e-3) put3(sqrt(rr),n,i)*/
      /* error of the iteration */
      f=SQR(dd)/Sqr(omegap)*si[i].qqpol;
      Max(maxerr,f)
      err+=f; ierr++; } }

  if (ierr!=No.free_s) prt("DOUBLE CHECK FAILED %d %d",ierr,No.free_s);

  maxerr=sqrt(maxerr);
  err=sqrt(err/ierr);
  if (option('a') && option_v&8) {
    static int first=1;
    static double lastmaxerr=1,lasterr=1;
    double ratio,maxratio;

    if (first) {
#  ifdef LAMBDA
      prt("  err_max     err_std     rate_max    rate_std     extrap_1     extrap_10");
#  else /*# LAMBDA */
      prt("omegap A      maxr  -> maxr  mol[site]  err_max  err_std  rate_max  rate_std");
#  endif /*#!LAMBDA */
      first=0; }
#  ifdef LAMBDA
    {
#    define QLAG 10
      static double q[3*QLAG];
      int i;
      double Q,QL,dq,mmm;
      double sum[3];

      for (i=3*QLAG-1; i; i--) q[i]=q[i-1];
      q[0]=err/lasterr;

      Q=q[1]-q[2];
      if (Q) Q=(dq=q[0]-q[1])/Q; else dq=0;
      loop (i,0,3) sum[i]=0;
      loop (i,0,3*QLAG) sum[i/QLAG]+=q[i];
      QL=sum[1]-sum[2];
      if (QL!=0 && (sum[0]-sum[1])/QL>=0)
        QL=pow((sum[0]-sum[1])/QL,1./QLAG);
      else
        QL=1;
      if (Q!=1) maxratio=q[0]+dq/(1-Q); else maxratio=0;
      if (QL!=1) mmm=q[0]+dq/(1-QL); else mmm=0;

      if (lastmaxerr) ratio=maxerr/lastmaxerr; else ratio=0;
      prt("%11.8f %11.8f %11.8f %13.10f %12.9f %12.9f",
          maxerr,err,
          ratio,q[0],
          maxratio,mmm);
      }
#  else /*# LAMBDA */
    if (lastmaxerr) maxratio=maxerr/lastmaxerr; else lastmaxerr=0;
    if (lasterr) ratio=err/lasterr; else lasterr=0;
    prt("%4.2f %.7f->%.7f %d[%d] %.5g %.5g %.7f %.8f",
        omegap, /* option('@') removed */
        sqrt(maxrrold),sqrt(maxrr),nmax,imax,
        maxerr,err,
        maxratio,ratio);
#  endif /*#!LAMBDA */
    lastmaxerr=maxerr,lasterr=err; }

  /* WARNING: En.self is added to En.pot outside forces() and selffield() */
  En.self/=2;

#  ifdef LAMBDA
  if (epsp<=0) {
    /* SPECIAL: 26 files for analysis by program lambda */
    static int iter,rev;
    static double maxerr0;

    /* note that xxx.z contains maxerr ! */
    loop (i,0,25)
      if (xxx[i]) fprintf(xxx[i],"%.16f\n",polarrof(cfg+i,A->rp)[LAMBDA][0]);
    if (xxx[25]) fprintf(xxx[25],"%.16f\n",maxerr);

    if (iter++>6 && maxerr/maxerr0>0.999) rev++;
    /*!!! wait for sig only */

    if (sig) {
      double q=10;

      if (sig==2) {
        fclose(out);
        exit(0); }
      if (sig==5) q=.1;
      loop (n,FROM,No.N) {
        mn=cfg+n;
        sp=mn->sp;
        ns=mn->ns;
        si=spec[sp]->si;
        rpol=polarrof(mn,A->rp);
        loop (i,0,ns) if (si[i].alpha_qq) VO(rpol[i],*=q) }
      prt("rescaled, sig=0 again");
      sig=0; }

#    if LAMBDA>=0
    if (rev>2 || sig) {
      /* finishing */
      double err[2],maxerr[2];
      int k;

      loop (k,0,2) err[k]=maxerr[k]=0;
      loop (n,FROM,No.N) {
        vector *twor[2];

        mn=cfg+n;
        ns=mn->ns;
        sp=mn->sp;
        si=spec[sp]->si;
        rpol=polarrof(mn,A->rp);
        loop (k,0,2)
          twor[k]=polarrof(mn,twofirst[k]->rp);
        loop (i,0,ns) {
          loop (k,0,2) {
            VVV(dd,=twor[k][i],-rpol[i])
            f=SQR(dd)*si[i].qqpol;
            Max(maxerr[k],f)
            err[k]+=f; } } }

      loop (k,0,2)
        prt("%s: maxerr=%.5f err=%.5f",k?"corr":"pred",
            sqrt(maxerr[k]),sqrt(err[k]/ierr));

      loop (i,0,26) if (xxx[i]) fclose(xxx[i]);
      exit(0); }
#    endif /*# LAMBDA>=0 */
    maxerr0=maxerr; }
#  endif /*# LAMBDA */

  if (t!=oldt) {
   /* here we record the errors of the predictor (in the 1st iteration) */
    StaSet(0,2,2,lag.n);
    StaAdd("polar one-step maxerr",En.polmaxerr=maxerr);
    StaAdd("polar one-step stderr",En.polstderr=err);
    oldt=t;
    nit=nbad=0; }

  nit++;

  if (nit>3) {
    if (maxerr>=maxerrold) nbad++;
    else nbad=0; }

#  if LAMBDA
  if (epsp>0)
#  endif /*# LAMBDA */
  if (nbad>2 || nit>maxitp) {
    static int origmaxitp;
    static double dmax[10];
    int nmax[10],imax[10],im;

    if (!origmaxitp) origmaxitp=maxitp;

    loop (im,0,10) dmax[im]=0;
    loop (n,FROM,No.N) {
      mn=cfg+n;
      ns=mn->ns;
      sp=mn->sp;
      si=spec[sp]->si;
      rpol=polarrof(mn,A->rp);
#  define ND 10
      loop (i,0,ns) {
        double d=SQR(rpol[i]);

        if (d>dmax[ND-1]) loop (im,0,ND) if (d>dmax[im]) {
          dmax[im]=d;
          imax[im]=i;
          nmax[im]=n;
          break; } } }

    prt("mol site |rpol| (%d worst)",ND);
    loop (im,0,ND) prt("%3d %2d %.7f",nmax[im],imax[im],dmax[im]);
#  undef ND

    WARNING(("selffield: %d iterations, %d consecutive bad\n\
*** (this may be OK with small epspx, esp. at start)\n\
*** try to increase maxitp",nit,nbad))
    if (!(option_v&8)) {
      static int pass;
      if (pass++) {
        prt("Bad for the second time - verbose output (-v8) set");
        option_v|=8; } }
    prt("maxitp increased to %d",maxitp+=maxitp);
    if (nbad>5 || nit>origmaxitp*10) ERROR(("selffield: too bad - stop")) }

  if (maxerr<epsp) {
    if (epsp==epspx) {
      StaAdd("selffieldx iter",nit);
      StaAdd("selffieldx maxr",sqrt(maxrr)); }
    else {
      StaAdd("selffield iter",En.polnit=nit);
      StaAdd("selffield maxr",En.polmaxdr=sqrt(maxrr)); }
    if (nit>1) StaAdd("selffield rate",maxerr/maxerrold);
    return 1; }
  else {
    maxerrold=maxerr;
    return 0; }
}

void mechpolar(ToIntPtr B,ToIntPtr A) /************************** mechpolar */
/*
  mechanical model of polarizable dipoles: just rhs
  NOTE: this was created as a simplified copy of selffield(),
        to be rewritten so that selffield() calls mechpolar()...
*/
{
  molecule_t *mn;
  siteinfo_t *si;
  int ns,sp,n,i;
  vector *fc,*fpol,*rpol,dr,dd;
  double rr,maxrrold=0,f;
#  ifndef FREEBC
  double byerd;
#  endif /*# FREEBC */

#  if POLAR&30
  ERROR(("mechpolar: POLAR=%d not supported",POLAR))
#  endif /*# POLAR&30 */

  zeroEn();

  forces(B,A);

  loop (n,FROM,No.N) {
    mn=cfg+n;
    ns=mn->ns;
    sp=mn->sp;
    si=spec[sp]->si;
    rpol=polarrof(mn,A->rp);
    fc=rof(mn,B->rp);
    fpol=polarrof(mn,B->rp);
    loop (i,0,ns) {
      VVV(dd,=dr,=rpol[i])
      rr=SQR(dr);
      Max(maxrrold,rr)
      if (rr>=Sqr(el.minqq))
        ERROR(("POLAR: %d[%d] aux charge dist = %g",n,i,sqrt(rr)))

      /* NOTE (9/2001): the virial of force equals -2*self energy and
         is treated separately because of the Ewald corrections */

#  if defined(COULOMB) && COULOMB<0
      /* Ewald correction for dipole self-term */
      f=si[i].charge*si[i].chargepol;
      En.el+=f*exacterud_sqrt(rr,&byerd);
      f*=byerd;
      VVO(fc[i],-=dr,*=f) VV(fpol[i],+=dr) /* note sign: dr=pol-center */
#  endif /*? FREEBC */ /*# defined(COULOMB) && COULOMB<0 */

      /* forces passed to the central atom */
      VV(fc[i],+=fpol[i]) } }
}
#endif /*# POLAR */

#ifdef SHEAR
void Shear(ToIntPtr B, ToIntPtr A, ToIntPtr VH,double H) /************ Shear */
/*
  (1) adds forces f_i[x] = f_i[y] = m_i*Kf*cos(2 PI z_i/L[z])
      where Kf=Cf/sqrt(2)  (because forces are in x and y directions)
  (2) calculates sumfc = SUM_i f_i[x] cos(2 PI z_i/L[z])
  (3) calculates sumpc = SUM_i m_i (v_i[x]+v_i[y]) cos(2 PI z_i/L[z])
      and Kv = sqrt(2)*Cv = sumpc*Kf/sumfc
  (4) records  num = sumfc*No.mass*L[2]/(2*PI*PI*L[0]*L[1])
      and      den = sumpc
      Then, <viscosity> = <num>/<den>
  (5) also Cv=sumpc/sumfc*Cf/2 is recorded in the convergence profile
      (amplitude of the velocity profile)
  (6) Twice the systematic kinetic energy contribution from the flow is
      En.kinshear = Kv^2*sumfc/2/Kf

  NOTES:
  * VH is a product of velocities and H (normally H=1, for Shake H=h)
  * For Shake, the velocities are by h/2 delayed (does not matter here)
  * Shear should be included before any thermostat (the systematic term
      En.Ekinshear has to be subtracted form the total [and
      intermolecular] energy)
*/
{
  molecule_t *mn;
  siteinfo_t *si;
  int ns,sp,n,i;
  vector *r,*f,*v;
  double
    cosz,m,fx,
    rho,
    phi,Kf,
    aux,
    sumfc,sumpc,sumfx;
  static int pass;

  if (shear==0) return;
  if (shear<0) ERROR(("negative shear not supported"))

  phi=2*PI/box.L[2];
  if (fabs(box.L[1]/box.L[0]-1)>1e-6)
    ERROR(("Shear implementation limitation: L[0]!=L[1]"))

  /* eta*dT/dt/rho^2 in program units: */
  rho=No.mass*rhounit/(box.L[0]*box.L[1]*box.L[2]);
  aux=shear/Sqr(rho)*(massunit/Pow5(lengthunit)*Sqr(timeunit));

  /* Kf=Cf/sqrt(2) because in x and y directions */
  Kf=PI*sqrt(2*aux*No.f/(Cub(box.L[2])*box.L[0]*box.L[1]));

  if (pass==0) {
    double Cf=Kf*sqrt(2.);
    double viscunit=Punit*timeunit;

    pass++;
    put3(shear,Kf,Cf);
    put2(viscunit,No.mass);
    put2(box.L[2],box.L[0]*box.L[1]*box.L[2]);
    prt("expected heating rate in prog.units [K/ps, etc.]: dE/dt = %g/eta",
        sqr(Cf*No.mass*box.L[2]/4/PI)/(box.L[0]*box.L[1]*box.L[2])); }

  /*
     NOTE to the estimate of Cf [Kf]: we use the `kinetic' estimate of
     heat capacity = No.f/2 (No.f would be probably better)
     because the same estimate is used in friction thermostats;
     (the time-scaling errors cancel)
  */

  sumfc=sumpc=sumfx=0;

  /* shear force + measurements */
  loop (n,FROM,No.N) {
    mn=cfg+n;
    ns=mn->ns;
    sp=mn->sp;
    si=spec[sp]->si;
    f=rof(mn,B->rp);
    r=rof(mn,A->rp);
    v=rof(mn,VH->rp);
    loop (i,0,ns) {
      m=si[i].mass;
      cosz=cos(phi*r[i][2]);
      fx=Kf*(m*cosz);
      sumfx+=fx;
      f[i][0]+=fx; f[i][1]+=fx;
      sumfc+=fx*cosz;
      sumpc+=(v[i][0]+v[i][1])*(m*cosz); } }

  /* correction to total acceleration in x,y directions
     (exact momentum conservation) */
  fx=sumfx/No.mass;
  loop (n,FROM,No.N) {
    mn=cfg+n;
    ns=mn->ns;
    sp=mn->sp;
    si=spec[sp]->si;
    f=rof(mn,B->rp);
    loop (i,0,ns) {
      m=si[i].mass;
      f[i][0]-=fx*m; f[i][1]-=fx*m; } }

  sumpc/=H;

  if (measure) {
    StaSet(0,2,2,lag.n);
    StaAdd("shear:num",No.mass*box.L[2]/(2*PI*PI*box.L[0]*box.L[1])*sumfc);
    StaAdd("shear:den",sumpc); }
  En.Cv=sumpc/sumfc*Kf*0.7071067811865476;
  En.kinshear=Sqr(En.Cv)*sumfc/Kf; /* twice! */

  /* kinetic T contribution equivalent to global cos-like velocity profile */
  StaSet(0,2,2,0);
  StaAdd("Tkinshear",En.kinshear/No.f);
}
#endif /*# SHEAR */

struct constrd_s constrd;
int rescale=15;        /* rescaling mode; 7=xyz, 8=CM */

void measureP(int pass) /***************************************** measureP */
/*
   Measures pressure and via selected diagonal terms of the pressure tensor
   using box rescaling (virtual volume change) in selected coordinates
   (see constrd.mode).  If densprof.slab&2 (=constrd.mode&RESCALE_SLAB),
   then x,y,z are rescaled so that volume does not change (cf. Gloor GJ,
   Jackson G, Blas FJ, de Miguel E: "Test-area simulation method for the
   direct determination of the interfacial tension of systems with continuous
   or discontinuous potentials", JCP, 123, 134703 (2005))
   POLAR: polar precision epspx must be small because of numerical derivative
     taken; also measures error of the self-field
   NOTES:
   - constrd.dV is in units of V (relative change)
   - uses two evaluations of forces (sparing one would mean to include this
     code to the place where the forces are normally calculated or exporting
     them which would be mess)
   - must work in 2 passes because Ekin need not be known while pass 1:
     measureP(1) : when positions are known
     measureP(2) : when Ekin is known
*/
{
  static ToIntPtr locA=NULL,locB,locRun,locEx;
  static int warn;
#ifdef POLAR
  int i,n;
  double Polstderr,Polmaxerr;
  FILE *frun=NULL,*fex=NULL,*ferr=NULL;
#endif /*# POLAR */
  static double Pkin,vir;
  double Ekin,Ep,Em;
  En_t En0;
  double V=box.L[0]*box.L[1]*box.L[2];
  vector L0;

  VV(L0,=box.L)

#ifdef DEBUG
  En_t Enp;
#endif /*# DEBUG */

  if (!constrd.dV) return;

  box.cq=Sqr(box.cutoff);

  if (pass==1) {
    /* numerical derivative dU/dV */

    if (No.c && !(rescale&RESCALE_CM) && !warn++)
      WARNING(("%d constrained bonds and atom-based rescaling:\n\
*** pressure by virtual volume change will be wrong\n\
*** - for small molecules use dV<0 (CM-based rescaling)\n\
*** - for large molecules use -u9999 (vibrating bonds) and check the thermostat\n\
*** note also that rescale=15 (\"CM\") does not affect virtual volume change",No.c))

    En0=En;
    VV(L0,=box.L)

    if (!locA) {
      sdsalloc(locA,a[0]->size);
      sdsalloc(locRun,a[0]->size); /* polar part would be enough...*/
      sdsalloc(locEx,a[0]->size); /* polar part would be enough...*/
      sdsalloc(locB,a[0]->size); }

    sdszero(locB);
    sdscopy(locA,a[0]);
    sdscopy(locRun,a[0]); /* polar part would be enough...*/

    rescalecfg(locA,constrd.mode,exp(constrd.dV/constrd.ncoord),0,0);

#ifdef POLAR
    do { /* iterate self-field */
#endif /*# POLAR */
      zeroEn();
      forces(locB,locA);
#ifdef POLAR
      } while (!selffield(locB,locA,epspx));

    Ep=En.pot+En.el+En.self;
    sdscopy(locEx,locA); /* polar part would be enough...*/
#else /*# POLAR */
    Ep=En.pot+En.el;
#endif /*#!POLAR */

#ifdef DEBUG
    Enp=En;
#endif /*# DEBUG */
    rescalecfg(locA,constrd.mode,exp(-2*constrd.dV/constrd.ncoord),0,0);

#ifdef POLAR
    do { /* iterate self-field */
#endif /*# POLAR */
      zeroEn();
      forces(locB,locA);
#ifdef POLAR
      } while (!selffield(locB,locA,epspx));
    Em=En.pot+En.el+En.self;
    Polmaxerr=Polstderr=0;

    if (option('v')&32) {
      char *fmode=init<2 || (init_append&2) ? "at" : "wt";

      frun=fopen(Fn("run.pol"),fmode);
      fex=fopen(Fn("ex.pol"),fmode);
      ferr=fopen(Fn("err.pol"),fmode);
      fprintf(frun,"\
#  t=%.4f  Running induced dipole moments in program units (0.01175014D)\n\
#  dip_x              dip_y              dip_z               |dip|\n",t);
      fprintf(fex,"\
#  t=%.4f  Iterated induced dipole moments in program units (0.01175014D)\n\
#  dip_x              dip_y              dip_z               |dip|\n",t);
      fprintf(ferr,"\
#  t=%.4f  Errors in induced dipole moments in program units (0.01175014D)\n\
#  dip_x              dip_y              dip_z               |dip|\n",t); }

    loop (n,FROM,No.N) {
      molecule_t *mn=cfg+n;
      int ns=mn->ns;
      siteinfo_t *si=spec[mn->sp]->si;
      vector *rpolRun=polarrof(mn,locRun->rp);
      vector *rpolEx=polarrof(mn,locEx->rp);
      vector *rpolA=polarrof(mn,locA->rp);
      vector dd;
      double f;

      loop (i,0,ns) if (si[i].chargepol) {
        VV(rpolEx[i],+=rpolA[i])
        VO(rpolEx[i],/=2)
        VVV(dd,=rpolRun[i],-rpolEx[i])
        f=SQR(dd)*si[i].qqpol;
        if (frun) fprintf(frun,"%18.12f %18.12f %18.12f  %18.12f\n",
                          rpolRun[i][0]*si[i].chargepol,
                          rpolRun[i][1]*si[i].chargepol,
                          rpolRun[i][2]*si[i].chargepol,
                          sqrt(SQR(rpolRun[i])*si[i].qqpol));
        if (fex) fprintf(fex,"%18.12f %18.12f %18.12f  %18.12f\n",
                         rpolEx[i][0]*si[i].chargepol,
                         rpolEx[i][1]*si[i].chargepol,
                         rpolEx[i][2]*si[i].chargepol,
                         sqrt(SQR(rpolEx[i])*si[i].qqpol));
        if (ferr) fprintf(ferr,"%18.12f %18.12f %18.12f  %18.12f\n",
                          dd[0]*si[i].chargepol,
                          dd[1]*si[i].chargepol,
                          dd[2]*si[i].chargepol,
                          sqrt(f));
        Polstderr+=f;
        Max(Polmaxerr,f) } }

    if (frun) { fprintf(frun,"\n"); fclose(frun); }
    if (fex) { fprintf(fex,"\n"); fclose(fex); }
    if (ferr) { fprintf(ferr,"\n"); fclose(ferr); }

    StaSet(0,lag.err,2,lag.n);
    Polstderr=sqrt(Polstderr/No.pol);
    Polmaxerr=sqrt(Polmaxerr);
    StaAdd("Polar stderr",Polstderr);
    StaAdd("Polar maxerr",Polmaxerr);
#else /*# POLAR */
    Em=En.pot+En.el;
#endif /*#!POLAR */

#ifdef DEBUG
    prt("~V      =%16.6f",V);
    prt("~r[0].x =%16.6f",rof(cfg,a[0]->rp)[0][0]);
    prt("~r[0].y =%16.6f",rof(cfg,a[0]->rp)[0][1]);
    prt("~r[0].z =%16.6f",rof(cfg,a[0]->rp)[0][2]);
    prt("~vir_dV =%16.6f",-(Ep-Em)/(2*constrd.dV));
    prt("~el_dV  =%16.6f",-(Enp.el-En.el)/(2*constrd.dV));
    prt("~pot_dV =%16.6f",-(Enp.pot-En.pot)/(2*constrd.dV));
    prt("~Ep     =%16.6f",Ep);
    prt("~Em     =%16.6f",Em);
    prt("~Enp.pot=%16.6f",Enp.pot);
    prt("~En.pot =%16.6f",En.pot);
    prt("~Enp.el =%16.6f",Enp.el);
    prt("~En.el  =%16.6f",En.el);
    prt("~En.vir =%16.6f",En.vir);
#  ifdef POLAR
    prt("~self_dV=%16.6f",-(En.self-Enp.self)/(2*constrd.dV));
 /*.....   prt("~both*3:=%16.6f",-(En.el+En.self-Enp.self-Enp.el)/(2*constrd.dV)*3);*/
    prt("~En.self=%16.6f",En.self);
#  endif /*# POLAR */
#endif /*# DEBUG */

    En=En0;
#ifndef FREEBC
    vir=-(Ep-Em)/(2*constrd.dV);
#endif /*# FREEBC */

#ifdef POLAR
    En.Polstderr=Polstderr;
    En.Polmaxerr=Polmaxerr;
#endif /*# POLAR */

    VV(box.L,=L0)
    VV(box.Lh,=0.5*box.L) }

  else if (pass==2) {
    /* kinetic contribution */
    if (constrd.mode & RESCALE_CM) Ekin=En.kin_tr;
    else Ekin=En.kin;

    if (constrd.mode & RESCALE_SLAB) Pkin=0;
    else Pkin=Ekin/(DIM*V)*Punit;

    En.PdV=Pkin+vir/V*Punit+constrd.Pcorr;
    StaSet(0,lag.err,2,lag.n);

    StaAdd(constrd.PdVname, En.PdV);
    StaSet(0,lag.err,2,2);
    StaAdd("PdVkin [Pa]",Pkin);
    StaAdd("dVvir",vir);
    StaAdd("PdVvir [Pa]",vir/V*Punit);

#ifdef DEBUG
    prt("~En.PdV =%16.6f",En.PdV);
    prt("~PdVkin =%16.6f",Pkin);
    prt("~dVvir  =%16.6f",vir);
    prt("~PdVvir =%16.6f",vir/V*Punit);
#endif /*# DEBUG */
    }
  else
    Error("measureP: bad pass");
}
