/* 
   Ewald pass 2 (calculate forces, energy, pressure tensor)
   #included from ewald.c 
*/
/* NB: Energy, PVIR, Qk are in the PARALLEL=1|3 version macros 
   referring to thread-based instances
*/

  loopto (kx,0,kxmax) {
    /* NOTE: 
       calculations of Qtab are unnecessarily repeated in the 2nd pass,
       however, storing all the factors would require a lot of memory,
       namely [ikk] indexed array of Qtab's */
    if (kx) iqeloop(ex) CxE(Qtab[iqt].x)
    iqloop Qtab[iqt].xy = Qtab[iqt].xY = Qtab[iqt].x;

    /* kymax=(int)(L[1]*sqrt(Kk-Sqr(kx/L[0]))); */
    kymax=ktab[kx].kymax;

    loopto (ky,0,kymax) {
      if (ky) iqeloop(ey) { CxE(Qtab[iqt].xy) CxiE(Qtab[iqt].xY) }

      iqloop {
        Qtab[iqt].xyz = Qtab[iqt].xyZ = Qtab[iqt].xy;
        Qtab[iqt].xYz = Qtab[iqt].xYZ = Qtab[iqt].xY; }

      /* kzmax=(int)(L[2]*sqrt(Kk-Sqr(ky/L[1])-Sqr(kx/L[0]))); */
      kzmax=ktab[kx].kzmax[ky];
      loopto (kz,!(kx+ky),kzmax) {
        if (kz) iqeloop(ez) {
          CxE(Qtab[iqt].xyz) CxiE(Qtab[iqt].xyZ)
          CxE(Qtab[iqt].xYz) CxiE(Qtab[iqt].xYZ) }

        if (el.alpha>=0) {
                  /********************************/
                  /**    pass 2 - normal Ewald   **/
                  /********************************/
          e=ekk[ikk]; 

#if PRESSURETENSOR&PT_VIR
/* to make more tables?...*/
	  ek[0]=e*(kx/box.L[0]); ek[1]=e*(ky/box.L[1]); ek[2]=e*(kz/box.L[2]);
	  kk=Sqr(kx/box.L[0])+Sqr(ky/box.L[1])+Sqr(kz/box.L[2]);
	  ptf=1/kk-PIaq;
#define CALCULATEPRESSURETENSOR \
          PVIR[0]+=E1*ptf*Sqr(kx/box.L[0]); \
          PVIR[1]+=E1*ptf*Sqr(ky/box.L[1]); \
          PVIR[2]+=E1*ptf*Sqr(kz/box.L[2]); 
#define ASSIGN_E1 E1=
#else
#define CALCULATEPRESSURETENSOR /* empty */
#define ASSIGN_E1 /* empty */
	  ek[0]=e*kx/box.L[0]; ek[1]=e*ky/box.L[1]; ek[2]=e*kz/box.L[2];
#endif

          switch ((kx>0) + (ky>0) + (kz>0)) {

/* +++ */   case 3:
              A=Qk[iQ++]; B=Qk[iQ++]; C=Qk[iQ++]; D=Qk[iQ++];
              Energy += ASSIGN_E1 (Csqr(A)+Csqr(B)+Csqr(C)+Csqr(D))*e;
              CALCULATEPRESSURETENSOR
#if PRESSURETENSOR&PT_OFF
              PVIR[3]+=e*ptf*(Csqr(A)-Csqr(B)-Csqr(C)+Csqr(D))*(ky/box.L[1])*(kz/box.L[2]);
              PVIR[4]+=e*ptf*(Csqr(A)-Csqr(B)+Csqr(C)-Csqr(D))*(kx/box.L[0])*(kz/box.L[2]);
              PVIR[5]+=e*ptf*(Csqr(A)+Csqr(B)-Csqr(C)-Csqr(D))*(kx/box.L[0])*(ky/box.L[1]);
#endif          
              iqfloop {
                FA = A.re*Qtab[iqt].xyz.im - A.im*Qtab[iqt].xyz.re; /* +++ */
                FB = B.re*Qtab[iqt].xyZ.im - B.im*Qtab[iqt].xyZ.re; /* ++- */
                FC = C.re*Qtab[iqt].xYz.im - C.im*Qtab[iqt].xYz.re; /* +-+ */
                FD = D.re*Qtab[iqt].xYZ.im - D.im*Qtab[iqt].xYZ.re; /* +-- */
                F[iqt][0] += ek[0]*((c=FA+FB)+(e=FC+FD));
                F[iqt][1] += ek[1]*(c-e);
                F[iqt][2] += ek[2]*(FA-FB+FC-FD); }
              break;

/* 0++ */
/* +0+ */
/* ++0 */   case 2:
              A=Qk[iQ++]; B=Qk[iQ++];
              Energy += ASSIGN_E1 (Csqr(A)+Csqr(B))*e;
              CALCULATEPRESSURETENSOR
/* ..+ */     if (kz) {
#if PRESSURETENSOR&PT_OFF
                PVIR[3]+=e*ptf*(Csqr(A)-Csqr(B))*(ky/box.L[1])*(kz/box.L[2]);
                PVIR[4]+=e*ptf*(Csqr(A)-Csqr(B))*(kx/box.L[0])*(kz/box.L[2]);
                PVIR[5]+=e*ptf*(Csqr(A)+Csqr(B))*(kx/box.L[0])*(ky/box.L[1]);
#endif          
                iqfloop {  
                  FA = A.re*Qtab[iqt].xyz.im - A.im*Qtab[iqt].xyz.re; /* ..+ */
                  FB = B.re*Qtab[iqt].xyZ.im - B.im*Qtab[iqt].xyZ.re; /* ..- */
                  if (kx) F[iqt][0] += ek[0]*(FA+FB);
                  else    F[iqt][1] += ek[1]*(FA+FB);
                          F[iqt][2] += ek[2]*(FA-FB); } }
/* ++0 */     else {
#if PRESSURETENSOR&PT_OFF
                PVIR[3]+=e*ptf*(Csqr(A)-Csqr(B))*(ky/box.L[1])*(kz/box.L[2]);
                PVIR[4]+=e*ptf*(Csqr(A)+Csqr(B))*(kx/box.L[0])*(kz/box.L[2]);
                PVIR[5]+=e*ptf*(Csqr(A)-Csqr(B))*(kx/box.L[0])*(ky/box.L[1]);
#endif
                iqfloop {
                  FA = A.re*Qtab[iqt].xyz.im - A.im*Qtab[iqt].xyz.re; /* .+0 */
                  FB = B.re*Qtab[iqt].xYz.im - B.im*Qtab[iqt].xYz.re; /* .-0 */
                  F[iqt][0] += ek[0]*(FA+FB);
                  F[iqt][1] += ek[1]*(FA-FB); } }
              break;

/* +00 */
/* 0+0 */
/* 00+ */   case 1:
              A=Qk[iQ++];
              Energy += ASSIGN_E1 Csqr(A)*e;
              CALCULATEPRESSURETENSOR
#if PRESSURETENSOR&PT_OFF
              PVIR[3]+=e*ptf*Csqr(A)*(ky/box.L[1])*(kz/box.L[2]);
              PVIR[4]+=e*ptf*Csqr(A)*(kx/box.L[0])*(kz/box.L[2]);
              PVIR[5]+=e*ptf*Csqr(A)*(kx/box.L[0])*(ky/box.L[1]);
#endif
              iqfloop {
                FA = A.re*Qtab[iqt].xyz.im - A.im*Qtab[iqt].xyz.re;
                if (kx)      F[iqt][0] += ek[0]*FA;
                else if (ky) F[iqt][1] += ek[1]*FA;
                else         F[iqt][2] += ek[2]*FA; }
              break;

/* ??? */   default: Error("Ewald switch"); } }

        else /* alpha<0 */ {
              /***********************************/
              /**    pass 2 - structure factor  **/
              /***********************************/
#if PARALLEL
          Error("PARALLEL structure factor not supported");
#endif

          if (!iscube())
	    ERROR(("sorry, structure factor implemented for a cube only"))

          jkk=kx*kx+ky*ky+kz*kz; 
          if (jkk>Sqr(el.kappa*box.L[2])) ERROR(("jkk>K^2"))

          switch ((kx>0) + (ky>0) + (kz>0)) {

/* +++ */   case 3:
              A=Qk[iQ++]; B=Qk[iQ++]; C=Qk[iQ++]; D=Qk[iQ++];
              sf[jkk].q += Csqr(A)+Csqr(B)+Csqr(C)+Csqr(D);
              sf[jkk].nk += 4;
              break;

/* 0++ */
/* +0+ */
/* ++0 */   case 2:
              A=Qk[iQ++]; B=Qk[iQ++];
              sf[jkk].q += Csqr(A)+Csqr(B);
              sf[jkk].nk += 2;
              break;

/* +00 */
/* 0+0 */
/* 00+ */   case 1:
              A=Qk[iQ++];
              sf[jkk].q += Csqr(A);
              sf[jkk].nk++;
              break;

/* ??? */   default: Error("Ewald switch"); } }

        ikk++;
        } } } /* kz,ky,kx */

  if (iQ!=Q->N) ERROR(("Ewald: iQ=%d != Q->N=%d; L=(%g %g %g)",
                               iQ,      Q->N,    VARG(box.L)))
