/* 3D version */
/* 2.4k:
   NearestImage(RET,CQ):
     RET = used in return RET; empty arg in void functions
           RET=empty is the original NearestImage
           RET=0     is the original NearestImage0
     CQ  = squared cutoff, 
           CQ=box.cq with a few exceptions:
             pure LJ, where CQ=ss->C2q
             cluster analysis
   ENPVIR = En.Pvir (serial) or glob->Pvir (parallel)
*/

/*
  optimized add forces macro: call by ADDFORCES(f1,f2,f*dr)
  f1=forces to atom 1 (vector),
  f2=forces to atom 2 (vector),
  f = value of force/|dr|,
  dr=r1-r2 (vector)
  real fdr  must be known
*/
#define ADDFORCES(F1,F2,F,DR) { \
  F1[0] += fdr = F*DR[0]; F2[0] -= fdr; \
  F1[1] += fdr = F*DR[1]; F2[1] -= fdr; \
  F1[2] += fdr = F*DR[2]; F2[2] -= fdr; }

#if PRESSURETENSOR&PT_VIR
/* NOTE: now unnecessarily duplicated ENVIR+- */
#if PRESSURETENSOR&PT_OFF
#define MADDFORCES(F1,F2,F,DR) { \
  F1[0] += fdr = F*DR[0]; F2[0] -= fdr; \
  ENPVIR[0]+=DR[0]*fdr; \
  ENPVIR[5]+=DR[1]*fdr; \
  F1[1] += fdr = F*DR[1]; F2[1] -= fdr; \
  ENPVIR[1]+=DR[1]*fdr; \
  ENPVIR[3]+=DR[2]*fdr; \
  F1[2] += fdr = F*DR[2]; F2[2] -= fdr; \
  ENPVIR[2]+=DR[2]*fdr; \
  ENPVIR[4]+=DR[0]*fdr; }
#else
#define MADDFORCES(F1,F2,F,DR) { \
  F1[0] += fdr = F*DR[0]; F2[0] -= fdr; \
  ENPVIR[0]+=DR[0]*fdr; \
  F1[1] += fdr = F*DR[1]; F2[1] -= fdr; \
  ENPVIR[1]+=DR[1]*fdr; \
  F1[2] += fdr = F*DR[2]; F2[2] -= fdr; \
  ENPVIR[2]+=DR[2]*fdr; }
#endif
#else
#define MADDFORCES(F1,F2,F,DR) ADDFORCES(F1,F2,F,DR)
#endif


#ifdef GOLD /* special GOLD version */
#include "goldmac.h"
#else /* !GOLD */

#define NI_f(I)  dr[I]=r1[I]-r2[I];

#ifdef WALL /* plain WALL (slit pore), not GOLD */
#define NI2 NI_f
#else
#define NI2 NI
#endif

#ifdef FREEBC

#define NearestImage(RET,CQ) NI_f(0) NI_f(1) NI_f(2) rr=SQR(dr); if (rr<CQ)

#ifdef NIBC
#error cannot FREEBC && NIBC
#endif
      
#else /* periodic b.c. with Ewald */

#ifdef WORM
  /*** molecules of length up to 3/2*L are allowed ***/
#define NI(I) \
  dr[I]=r1[I]-r2[I]; \
  if      (dr[I]<-box.Lh[I]) { dr[I]+=box.L[I]; if (dr[I]<-box.Lh[I]) dr[I]+=box.L[I]; } \
  else if (dr[I]> box.Lh[I]) { dr[I]-=box.L[I]; if (dr[I]> box.Lh[I]) dr[I]-=box.L[I]; }
#else
/*** molecules of length up to L/2 are allowed ***/
#define NI(I) \
   dr[I]=r1[I]-r2[I]; if (dr[I]<-box.Lh[I]) dr[I]+=box.L[I]; else if (dr[I]>box.Lh[I]) dr[I]-=box.L[I];
#endif

#ifdef CUT /* efficient if cutoff is significantly less than half box size */
/* see also eldens.c  - to be improved !!! */
#define NearestImage(RET,CQ) \
  NI(0) if ((rr =dr[0]*dr[0])>=CQ) return RET; \
  NI(1) if ((rr+=dr[1]*dr[1])>=CQ) return RET; \
  NI2(2)     rr+=dr[2]*dr[2]; \
  if (rr<CQ)

#elif defined(NIBC) /* nearest-image b.c without spherical cutoff */
#define NearestImage(RET,CQ)  NI(0) NI(1) NI2(2) rr=SQR(dr);

#else /* standard periodic b.c. */
#define NearestImage(RET,CQ)  NI(0) NI(1) NI2(2) rr=SQR(dr); if (rr<CQ)
#endif

#endif /* !FREEBC */
#endif
