#define VERSION "March 2000"

#ifdef LINKCELL
#define METHOD "link-cell"
#ifdef FREEBC
#error FREEBC and LINKCELL not supported
{This is to force stupid compilers to stop at error}
#endif
#else
#ifdef NIBC
#define METHOD "nearest image"
#elif FREEBC
#define METHOD "free b.c."
#elif TWODIM
#define METHOD "2D"
#else
#define METHOD ""
#endif
#endif

#ifdef FLOAT
#define HEADER METHOD VERSION "/float"
#else
#define HEADER METHOD VERSION
#endif

#ifndef HIT
#define SIG /* to enable the ^C handler */
#endif

#include "ground.h"
#define SDS
#include "alloc.h"
#include <time.h>

#include "options.h"
#define x badoption
int optionlist[32] =
/*`   a b c d  e f g h  i j k l m n o  p   q r s t u v w x y z { | }      */
{-9,100,0,4,1,-1,0,0,0,-9,0,0,0,2,0,0,-1,100,1,0,0,0,1,2,0,0,0,1,0,0,-9,0};
/*@   A B C D  E F G H  I J K L M N O  P   Q R S T U V W X Y Z [ \ ] ^  _ */
#undef x

volatile int sig;

#include "varfile.h"
#include "statics.h"
#include "linregr.h"

#include "rndgen.h"
#include "simglob.h"

#if defined(MPI) && PARALLEL==3
#include <mpi.h>
#include "formain.h"
#endif

#ifndef PARALLEL
#define PARALLEL 0
#endif

/* asksig() and SIG,HIT signal handling removed */

#if defined(Hamilton) == defined(Lagrange)
#error Only one of {Hamilton Lagrange} may be #defined
{This is to force stupid compilers to stop at error}
#endif

#include "units.h"
#include "simgear.h"
#include "simcg.h"

#include "interpot.h"
#include "simdef.h"

#include "mainlib.h"

/*** some additional global variables ***/

#ifdef SHAKE
int shake=0;
#endif
#ifdef CACHE
int by=32;        /* # of molecules in a subblock */
#endif
#ifdef LINKCELL
real over14=1.5;  /* factor by which the estimate of change of the max
                     1-2,1-3,1-4 distance is multiplied.
                     Must be >1; the higher, the safer
                     negative: also print distances and speeds
                     1.5 is pessimistic for proteins,
                     but just enough for water */
#endif
#if PARALLEL==1 || PARALLEL==3
int measureserial;
#endif

/*
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %                       Ewald summation & erfc                          %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

##ifdef EWALDLIKE
#include "ewald.h"
#include "erfc.h"
#endif

#include "simpot.h"
#include "siminit.h"
#ifdef DIHEDRAL
/* WARNING this is something very obsolete... */
#include "dihedral.h"
#endif
#include "simmeas.h"

#include "norm.h"
#include "simils.h"
#include "rhs.h"
#include "constrd.h"
#include "cputime.h"

#include "sfdx.h"

#ifdef MARK
#include "mark.h"
#endif

#ifdef SCALEDMOL
#include "scalepot.h"
#endif

void initcook(char *blendfile,double cut,double LJcutoff) /******** initcook */
{
int i;
char *mark;

out=stdout; in=stdin;

if (!blendfile) ERROR(("initcook: no blend file"))

if (option('p')<0) option('p')=option('m');
if (option('s')==1)
#ifdef DOS
  option('s')=31;
#else
  option('s')=256;
#endif

#if PARALLEL
if (option('i')==-9) option('i')=1+(!option('s'));
if (option('i')==0) WARNING(("option -i0 for PARALLEL"))
#else
if (option('i')==-9) option('i')=2*!option('s');
#endif

if (option('f')) {
  option('q')=0;
  prt("option -q0 set because of -f"); }

#ifdef POLAR
if (option('f')) {
  option('a')=0;
  prt("option -a0 set because of -f"); }

if (option('@')==-9) option('@')=option('m')<4?50:0;
#else
if (option('@')==-9) option('@')=0;
#endif
if (option('^')==-9) option('^')=option('m')<4?60:100;
option('n')=option('n')!=0;

#if PARALLEL && !defined(MPI)
{
char *env=getenv("MP_SET_NUMTHREADS");
if (!env) ERROR(("setenv MP_SET_NUMTHREADS required"))
#if PARALLEL==1
/* -\\0 means No.cell threads even if MP_SET_NUMTHREADS<No.cell */
if (option('\\')<0) option('\\')=atoi(env);
#else
/* will be partitioned to -\\threads, default=MP_SET_NUMTHREADS */
if (option('\\')<=0) option('\\')=atoi(env);
#endif
}
#endif

#ifdef Lagrange
#ifdef SHAKE
shake=option('m')==2;
if (shake) option('c')=0;
#else
if (option('m')<=2) Error("SHAKE not #defined and -m2");
#endif
#endif

if (option('c')==4) epsc=0.05;

#if 0
if (option('j')<0) {
  noinitvel= -option('j');
  option('j')=0; }
else
  noinitvel=0;
#endif

initscroll(option('s')*1024);

out = option('s') ? stdout : fopen("initcook.prt","w");
if (out==NULL) Error("open initcook.prt");

prts("*** " PROJECT " Version " HEADER " *** (c) J.Kolafa 1991- ***");

#ifdef Lagrange
prts_("*** Lagrange");
#else
prts_("*** Hamilton");
#endif

#ifdef FREEBC
prts_(" *** FREEBC");
#endif

#ifdef POLAR
prts_(" *** POLAR");
#endif

#if PARALLEL
prt_(" *** PARALLEL=%d",PARALLEL);
#endif

#ifdef REDUNITS
prts_(" *** REDUNITS");
#endif

{
int Int=1;
char *end=(char*)&Int;

prt(" *** %s endian s%d i%d l%d p%d ***",
  *end==1?"little":*end==0?"BIG":"?",
  sizeof(short),sizeof(int),sizeof(long),sizeof(void*));
}

prt_("%s ",blendfile);
_n

loop (i,0,32) {
  if (i==16) _n 
  prtoption(i,optionlist) }
_n

#if PARALLEL
if (option('i')==0) WARNING(("option -i0 for PARALLEL"))
#if PARALLEL==2
if (option('\\')!=2 && option('\\')!=1)
  Error("number of threads must be 1 or 2");
#else
if (option('\\')<0) Error("bad option -\\");
#endif
#endif

rndinit(7L/*low quality enough*/,option('z'));

cutoff=cut;
cq=cut*cut;

#ifdef POLAR
/* POLAR: omegap given by option -\^%, can be changed from data */
omegap=option('^')/100.;
#endif

/* //////////////////////// read blend file ///////////////////// */

#ifdef COOK
in=fopen(blendfile,"rt");
if (!in) Error("open blend file");
readblend();
fclose(in);
#else
convertsiteunits(); /* just writes info #ifdef REDUNITS */
#endif

#if defined(MPI) && PARALLEL==3
BroadcastInit(&narg,&arg);
#endif

/* 
   make a `configuration' with 1 molecule of all species 
   otherwise tables of molecules with N=0 are not initialized !!! 
*/
{
int sp;
loop (sp,0,nspec) spec[sp]->N=1;
}

checkranges(1);

initNo(); /* now we know everything to call it */

measure=1; /* to measure energy */

/* total charge check delayed to be eventually interactive */
{
int sp;
double tc;

loop (sp,0,nspec) {
  tc=spec[sp]->tc/electron;
  if (fabs(fmod(fabs(tc+0.5),1)-0.5) > 1e-5)
    WARNING(("\
charge = %g e of species %d is fractional\n\
(you may need blend -_, check also mol-file)",tc,sp)) }
}

#ifdef EWALDLIKE
#if SPLINE==-1
if (K<=0) WARNING(("K=%g is suspicious for Ewald",K))
#else
if (K) WARNING(("K=%g is suspicious for cutoff electrostatics",K))
#endif
#endif

GearOrder=option('m'); /* to be eventually removed and set GearOrder otherway */

#if PARALLEL==1 || PARALLEL==3
if (rdf.grid) {
  prt("WARNING: rdf not parallelized, all measurements serial (check noint)");
  measureserial++; }
#ifdef DIHHIST
if (dih.grid) {
  prt("WARNING: dih not parallelized, all measurements serial (check noint)");
  measureserial++; }
#endif
#endif

/* ////////////////// REPEATABLE INITIALIZATIONS /////////////////// */

ralloc(mark,1); /* !! all heap cleared by release(mark) !! */

#ifndef COOK
initspecies(); /* init spec + calculates No */
#endif

initcutoffL(); /* may re-set cutoff, needs No */

/*
  rdf measurements are embedded in site-site interactions 
  must be initialized to `empty; even if not used
*/
initrdf(0,LJcutoff,NULL);
/*.....initrdf(GRID,CUTOFF,SITE_SITE_FILE);*/

#if 0
  /* /////////////////// optional initializations ////////////////// */

#ifdef DIHHIST
  initdihhist(GRID); /* only histograms - control structures already inited*/
#endif

#ifdef HARMONICS
  /* very special */
  initharmonics(GRID,CUTOFF,T,rho);
#endif
#endif /* optional initializations */

/* init site-site interactions */
initss(LJcutoff);

/* assign table of pair potentials */
initpot();

/* init tables mapping configuration */
option('m')=0; /* will allocate a[0] (r) and a[1] (forces) */
initmolecules();

#if 0
  /* ////////////////////// CONFIGURATION INTIALIZATIONS /////////////////// */
  corr=cutcorr(LJcutoff); /* must always call because of counting rdf->npairs */

#if defined(TWODIM) || defined(FREEBC)
  corr=0; /* not supported */
#endif

  prt("rho = %.12g / V  (for rho in kg m^-3 and V in A^3)",rhounit*No.free_mass);

  /* for constraint dynamics */
  initgammas();

  /* ... reading or initializing  config, measurements... not included */
#endif
}

double energy1(int sp,vector *r)
{
molecule_t *m=cfg+sp;
vector *f=rof(m,a[1]->rp);

En.el=0;
copy(rof(m,a[0]->rp),r,m->ns*sizeof(vector));
return (*pot[sp][sp]) (f,f,m,m,a[0]->rp)+En.el;
}

double forces1(int sp,vector *r,vector **pf)
{
molecule_t *m=cfg+sp;
vector *f=rof(m,a[1]->rp);

*pf=f;

En.el=0;
copy(rof(m,a[0]->rp),r,m->ns*sizeof(vector));
memset(f,0,m->ns*sizeof(vector));
return (*pot[sp][sp]) (f,f,m,m,a[0]->rp)+En.el;
}

double energy2(int sp1,int sp2,vector *r1,vector *r2)
{
molecule_t *m1=cfg+sp1;
molecule_t *m2=cfg+sp2;
vector *f1=rof(m1,a[1]->rp);
vector *f2=rof(m2,a[1]->rp);

En.el=0;
copy(rof(m1,a[0]->rp),r1,m1->ns*sizeof(vector));
copy(rof(m2,a[0]->rp),r2,m2->ns*sizeof(vector));
return (*pot[sp1][sp2]) (f1,f2,m1,m2,a[0]->rp)+En.el;
}

double forces2(int sp1,int sp2,vector *r1,vector *r2,vector **pf1,vector **pf2)
{
molecule_t *m1=cfg+sp1;
molecule_t *m2=cfg+sp2;
vector *f1=rof(m1,a[1]->rp);
vector *f2=rof(m2,a[1]->rp);

*pf1=f1;
*pf2=f2;

En.el=0;
memset(f1,0,m1->ns*sizeof(vector));
copy(rof(m1,a[0]->rp),r1,m1->ns*sizeof(vector));
memset(f2,0,m2->ns*sizeof(vector));
copy(rof(m2,a[0]->rp),r2,m2->ns*sizeof(vector));
return (*pot[sp1][sp2]) (f1,f2,m1,m2,a[0]->rp)+En.el;
}

