/*
  Verlet step (for 1 molecule) with posible thermostats:
  - atom-based Andersen/Maxwell thermostat
  - CM-based Andersen/Maxwell thermostat
  - Nose with predicted velocity
  #included twice from constrd.c before two versions of SHAKE
  input:  p[i]     = force on atom i at time t
          r[i]     = position of atom i at time t
          r1[i]    = r[i](t)-r[i](t-h) = h*v(t-h/2)
          Vpred[0] = h*d logs(t)/dt, predicted
          vpred[i] = h*velocity of atom i, predicted
  output: p[i] = r[i](t+1) (before SHAKE)
          r1 is not changed

  Andersen/Maxwell thermostat for models with constraints is not recommended!
  (previous version without Nose = shakemax.c)
*/

  /* Nose-Hoover thermostat with predicted velocities */
  if (thermostat==T_NOSE) {
    loop (i,0,ns) {
      VVV(p[i],=hh*si[i].imass*p[i],-Vpred[0]*v[i])
      VVV(p[i],+=r[i],+r1[i]) } }

  /* Andersen/Maxwell thermostat */
  else if (prob) {
    if (CMbasedthermostat) {
      if (rnd()<prob) { /* center-of-mass-based thermostat */
        vector cmx;
        double im=1./spec[mn->sp]->mass;
        double sigma=sqrt(Thh*im);

        VO(cmx,=0)
        loop (i,0,ns) VV(cmx,-=si[i].mass*r1[i])
        VVO(cmx,=im*cmx,+rndgauss()*sigma)
        loop (i,0,ns) {
          VO(p[i],*=hhh*si[i].imass)
          VVVV(p[i],+=r[i],+r1[i],+cmx) } }
      else loop (i,0,ns) { /* no thermostat now */
        VO(p[i],*=hh*si[i].imass)
        VVV(p[i],+=r[i],+r1[i]) } }

    else loop (i,0,ns) /* atom-based thermostat */
      if (rnd()<prob) {
        double sigma=sqrt(Thh*si[i].imass);

        VO(p[i],*=hhh*si[i].imass)
        VVO(p[i],+=r[i],+rndgauss()*sigma) }
      else { /* no thermostat now */
        VO(p[i],*=hh*si[i].imass)
        VVV(p[i],+=r[i],+r1[i]) } }

  else /* no thermostat */
    loop (i,0,ns) {
      VO(p[i],*=hh*si[i].imass)
      VVV(p[i],+=r[i],+r1[i]) }
