/* \make cfg2plb

  6/2009 - bug fixed (asked for confirmation even if no file to overwrite)
*/

#define SDS
#include "ground.h"
#include "alloc.h"
#include "varfile.h"

typedef double vector[3];

int main(int narg,char **arg)
{
  FILE *f;
  float header[2];
  vector L;
  float Lf[3];
  char *fn;
  struct a_s {
    int  size;    /* whole struct in bytes */
    int  padd;    /* padded to 8 bytes */
    double logs;  /* log of the Nose variable s */
    vector rp[1]; /* contiguous array of all r[ns] (and p[ns]) */
    } *a;  
  int i,k,cfgkey,N,ns,NS,oldsize;
  int optionlist[32];
  double t,h,Entot,RvdW,zero;

  if (narg<2) {
    fprintf(stderr,"Convert cfg-file into plb-file. Call by:\n\
  cfg2plb NAME [NS]\n\
where\n\
  NAME.cfg will be converted into NAME.plb (one frame)\n\
  (asks for confirmation if NAME.plb would be overwritten)\n\
  NS is optional number of sites (to truncate the configuration)\n\
Limitations:\n\
  some information is lost on conversion\n\
  cannot handle cfg-files written in float or with reversed endianness\n\
See also:\n\
  plbinfo plb2asc asc2plb plbconv plb2plb plb2cfg\n");
    exit(0); }

  initscroll(0);
  
  VarOpen(fn=string("%s.cfg",arg[1]),"r");
  prt("reading %s",fn);
  VarRead(&cfgkey,sizeof(cfgkey));
  VarRead(optionlist,sizeof(optionlist));
  VarRead(&N,sizeof(N));
  put2(cfgkey,N)
  if (!(cfgkey&8)) ERROR(("sorry, old version of cfg-file is not supported"))
  if (VarFile.size!=8) ERROR(("cfg-file is not in double or has bad format"))
  VarRead(L,sizeof(L[0]));
  VarRead(&t,sizeof(t));
  VarRead(&h,sizeof(h));
  VarRead(&Entot,sizeof(Entot));
  put3(t,h,Entot)

  i=12-2;
  /* added in 2.4b */
  VarRead(&RvdW,sizeof(RvdW)); i--;

  while (i--) VarRead(&zero,sizeof(zero));
  VarRead(L+1,sizeof(L[1]));
  VarRead(L+2,sizeof(L[2]));
  putv(L)
  
  ns=(VarFile.size-sizeof(*a))/sizeof(vector)+1;
  if (cfgkey&2) {
    prt("POLAR configuration detected");
    ns/=2; }
  put(ns)
  if (ns<N) ERROR(("ns<N read from file"))
  if (narg>2) NS=atoi(arg[2]);
  else NS=ns;
  if (NS>ns) prt("WARNING: NS>ns specified, NS=ns=%d assumed",NS=ns);
  if (NS<ns) prt("WARNING: smaller NS=%d specified, configuration will be truncated",NS);
  
  sdsalloc(a,VarFile.size);
  oldsize=VarFile.size;
  VarRead(a,VarFile.size);
  VarClose();
  if (a->size!=oldsize) 
    ERROR(("size of SDS in file (%d) does not match VarFile.size (%d)",a->size,oldsize))

  f=fopen(fn=string("%s.plb",arg[1]),"r");
  if (f) {
    WARNING(("%s exists and is about to be overwritten",fn))
    fclose(f); }
            
  f=fopen(fn,"wb");
  if (!f) ERROR(("cannot open %s for writing",fn))

  header[0]=NS;
  header[1]=-3;

  if (fwrite(header,4,2,f)!=2) ERROR(("cannot write to %s",fn))
  loop (k,0,3) Lf[k]=L[k];
  if (fwrite(Lf,4,3,f)!=3) ERROR(("cannot write to %s",fn))
  loop (i,0,NS) {
    loop (k,0,3) Lf[k]=a->rp[i][k];
    if (fwrite(Lf,4,3,f)!=3) ERROR(("cannot write to %s",fn)) }
  fclose(f);

  prt("%s written",fn);

  return 0;
}
