/* cc -o plbreplicate plbreplicate.c
*/

#include "../gen/include.h"

typedef float vector[3];

int main(int narg,char **arg)
{
  FILE *f1,*f2;
  int ns,i,nx,ny,nz,ix,iy,iz;
  static float header[2];
  vector *r;
  vector L,newL;

  if (narg<5) {
    fprintf(stderr,"Replicate a cell. Call by:\n\
  plbrepl INPUT_FILE OUTPUT_FILE NX NY NZ\n\
INPUT_FILE is a plb-file with box info (new format)\n\
WARNING: not suitable as cook* input if the cell contains more species\n\
See also:\n\
  cook* -[XYZ   molcfg\n");
    exit(0); }

  if (!strcmp(arg[1],arg[2])) Error("INPUT_FILE = OUTPUT_FILE");

  f1=fopen(arg[1],"rb");
  if (f1==NULL) Error("no INPUT_FILE");

  f2=fopen(arg[2],"wb");
  if (f2==NULL) Error("cannot open OUTPUT_FILE");

  fread(header,sizeof(header),1,f1);
  ns=header[0];

  if (header[1]!=-3) Error("wrong header (variable box required), try plbconv");

  allocarray(r,ns);

  nx=atoi(arg[3]);
  ny=atoi(arg[4]);
  nz=atoi(arg[5]);
  
  header[0]=ns*nx*ny*nz;
  if (header[0]>16777216.||header[0]<1) Error("bad number of sites (max. 16777216)");
  
  fwrite(header,sizeof(header),1,f2);
  fread(L,sizeof(L),1,f1);
  newL[0]=L[0]*nx;
  newL[1]=L[1]*ny;
  newL[2]=L[2]*nz;
  if (ns!=fread(r,sizeof(vector),ns,f1)) Error("plb file truncated");

  fwrite(newL,sizeof(newL),1,f2);

  loop (ix,0,nx) 
    loop (iy,0,ny) 
      loop (iz,0,nz) 
        loop (i,0,ns) {
          vector newr;

          newr[0]=L[0]*ix+r[i][0];
          newr[1]=L[1]*iy+r[i][1];
          newr[2]=L[2]*iz+r[i][2]; 
          fwrite(newr,sizeof(newr),1,f2); }

  fclose(f2);
  fclose(f1);

  printf("! plbreplicate %s %s %d %d %d\n\
! N[0]=%.0f ! the total number of sites\n\
L[0]=%.7f\n\
L[1]=%.7f\n\
L[2]=%.7f\n",
         arg[1],arg[2],nx,ny,nz,
         header[0],newL[0],newL[1],newL[2]);


  return 0;
}
