/* gcc -o plbrot plbrot.c -lm
*/
#include "../gen/include.h"
typedef double real;
#include "../gen/vector3d.h"

int axis;

void rotate(float *r,double angle) /******************************** rotate */
{
  double sa=sin(angle), ca=cos(angle);
  double o[3][3];
  float rr[3];
  int i=(axis+1)%3,j=(i+1)%3,k=(j+1)%3,a;
 
  o[i][i]=ca; o[i][j]=-sa; o[i][k]=0;
  o[j][i]=sa; o[j][j]= ca; o[j][k]=0;
  o[k][i]= 0; o[k][j]=  0; o[k][k]=1;
 
  loop (a,0,3) rr[a]=SCAL(o[a],r);
 
  VV(r,=rr)
}

int main(int narg,char **arg)
{
  int i,f=0,ns1,ns2,varL=0;
  float r[3],center[3];
  float dr[3]={0,0,0};
  FILE *plb1,*plb2;
  double angle,dangle;
  int is3db=0;

  if (narg<6) {
    fprintf(stderr,"Rotate a plb or 3db-file. Call by:\n\
  plbrot INPUT.plb OUTPUT.plb AXIS ANGLE DANGLE [DX DY DZ]\n\
  plbrot INPUT.3db OUTPUT.3db AXIS ANGLE NS [DX DY DZ]\n\
Rotates INPUT.plb around axis AXIS={x,y,z} (1 char)\n\
by angle ANGLE+frame*DANGLE where frame is numbered from 0.\n\
Periodic boundary conditions: rotates around box center.\n\
The output plb-file is old format, L=0, center=(0,0,0)\n\
Angles are in degrees. Any other extension than .3db means a plb-file.\n\
Optional DX DY DZ = translation.\n\
Example:\n\
  plbrot cfg.plb rot.plb z 0 1\n");
    exit(0); }

  if (!strcmp(strend(arg[1])-4,".3db")) is3db++;
  if (!strcmp(strend(arg[2])-4,".3db")) is3db++;
  if (is3db==1) Error(".3db extensions do not match");

  plb1=fopen(arg[1],"rb");
  if (!plb1) Error(arg[1]);
  if (!strcmp(arg[1],arg[2])) Error("infile=outfile");
  plb2=fopen(arg[2],"wb");
  axis=arg[3][0]-'X';
  while (axis>3) axis-=16;
  if (axis<0 || axis>=3) Error("wrong axis");
  angle=atof(arg[4])*(PI/180);
  dangle=atof(arg[5])*(PI/180);
  ns1=ns2=atoi(arg[5]);

  if (narg>6) dr[0]=atof(arg[6]);
  if (narg>7) dr[1]=atof(arg[7]);
  if (narg>8) dr[2]=atof(arg[8]);

  if (!is3db) {
    fread(r,4,2,plb1);
    varL=r[1]<0;
    ns1=r[0]; }
  if (ns1<1 || ns1>16777216) Error("wrong ns - endian?");
  if (r[1]>0) center[0]=center[1]=center[2]=r[1]/2;
  r[1]=0; /* old output format */
  if (!is3db)
    fwrite(r,4,2,plb2);

  for (f=0;;f++) {
    double a=angle+dangle*f;
    
    if (varL) {
      if (3!=fread(r,4,3,plb1)) goto end;
      center[0]=r[0]/2;
      center[1]=r[1]/2;
      center[2]=r[2]/2; }
    
    loop (i,0,ns1) {
      if (3!=fread(r,4,3,plb1)) goto end;
      VV(r,-=center)
      rotate(r,a);
      VV(r,+=dr)
      if (3!=fwrite(r,4,3,plb2)) Error("cannot write"); } 
    printf("frame %d rotated by %f around %c\n",f,a,axis+'x'); }

 end: 
  fclose(plb1);
  fclose(plb2);

  return 0;
}
